# coding: utf-8
from __future__ import unicode_literals

from django.db import models

import django_tables2 as tables


def test_should_turn_email_address_into_hyperlink():
    class Table(tables.Table):
        email = tables.EmailColumn()

    table = Table([{'email': 'test@example.com'}])
    assert table.rows[0].get_cell('email') == '<a href="mailto:test@example.com">test@example.com</a>'


def test_should_render_default_for_blank():
    class Table(tables.Table):
        email = tables.EmailColumn(default='---')

    table = Table([{'email': ''}])
    assert table.rows[0].get_cell('email') == '---'


def test_should_be_used_for_emailfields():
    class EmailModel(models.Model):
        field = models.EmailField()

        class Meta:
            app_label = 'django_tables2_test'

    class Table(tables.Table):
        class Meta:
            model = EmailModel

    assert type(Table.base_columns['field']) == tables.EmailColumn


def test_text_should_be_overridable():
    class Table(tables.Table):
        email = tables.EmailColumn(text='@')

    table = Table([{'email': 'test@example.com'}])
    assert table.rows[0].get_cell('email') == '<a href="mailto:test@example.com">@</a>'


def test_value_returns_a_raw_value_without_html():
    class Table(tables.Table):
        col = tables.EmailColumn()

    table = Table([{'col': 'test@example.com'}])
    assert table.rows[0].get_cell_value('col') == 'test@example.com'



# coding: utf-8
import pytest
from django.core.exceptions import ImproperlyConfigured
from django.views.generic.base import TemplateView

import django_tables2 as tables

from .app.models import Person, Region
from .utils import build_request

MEMORY_DATA = [
    {'name': 'Queensland'},
    {'name': 'New South Wales'},
    {'name': 'Victoria'},
    {'name': 'Tasmania'}
]


class DispatchHookMixin(object):
    '''
    Returns a response *and* reference to the view.
    '''

    def dispatch(self, *args, **kwargs):
        return super(DispatchHookMixin, self).dispatch(*args, **kwargs), self


class SimpleTable(tables.Table):
    class Meta:
        model = Region


class SimpleView(DispatchHookMixin, tables.SingleTableView):
    table_class = SimpleTable
    model = Region  # required for ListView


class SimplePaginatedView(DispatchHookMixin, tables.SingleTableView):
    table_class = SimpleTable
    table_pagination = {'per_page': 1}
    model = Region  # required for ListView


@pytest.mark.django_db
def test_view_should_support_pagination_options():
    for region in MEMORY_DATA:
        Region.objects.create(name=region['name'])

    response, view = SimplePaginatedView.as_view()(build_request('/'))
    assert view.get_table().paginator.num_pages == len(MEMORY_DATA)
    assert view.get_table().paginator.per_page == 1


@pytest.mark.django_db
def test_view_should_support_default_pagination():
    class PaginateDefault(DispatchHookMixin, tables.SingleTableView):
        table_class = SimpleTable
        model = Region
        table_data = MEMORY_DATA

    response, view = PaginateDefault.as_view()(build_request('/'))
    table = view.get_table()
    assert table.paginator.per_page == 25
    assert len(table.page) == 4


@pytest.mark.django_db
def test_view_should_support_default_pagination_with_table_options():
    class Table(tables.Table):
        class Meta:
            model = Region
            per_page = 2

    class PaginateByDefinedOnView(DispatchHookMixin, tables.SingleTableView):
        table_class = Table
        model = Region
        table_data = MEMORY_DATA

    response, view = PaginateByDefinedOnView.as_view()(build_request('/'))
    table = view.get_table()
    assert table.paginator.per_page == 2
    assert len(table.page) == 2


@pytest.mark.django_db
def test_view_should_support_disabling_pagination_options():
    class SimpleNotPaginatedView(DispatchHookMixin, tables.SingleTableView):
        table_class = SimpleTable
        table_data = MEMORY_DATA
        table_pagination = False
        model = Region  # required for ListView

    response, view = SimpleNotPaginatedView.as_view()(build_request('/'))
    table = view.get_table()
    assert not hasattr(table, 'page')


@pytest.mark.django_db
def test_view_from_get_queryset():
    for region in MEMORY_DATA:
        Region.objects.create(name=region['name'])

    class GetQuerysetView(SimpleView):
        def get_queryset(self):
            return Region.objects.filter(name__startswith='Q')

    response, view = GetQuerysetView.as_view()(build_request('/'))
    table = view.get_table()

    assert len(table.rows) == 1
    assert table.rows[0].get_cell('name') == 'Queensland'


def test_should_raise_without_tableclass():
    class WithoutTableclassView(tables.SingleTableView):
        model = Region

    with pytest.raises(ImproperlyConfigured):
        WithoutTableclassView.as_view()(build_request('/'))


def test_should_support_explicit_table_data():
    class ExplicitDataView(SimplePaginatedView):
        table_data = MEMORY_DATA

    response, view = ExplicitDataView.as_view()(build_request('/'))
    assert view.get_table().paginator.num_pages == len(MEMORY_DATA)


@pytest.mark.django_db
def test_paginate_by_on_view_class():
    Region.objects.create(name='Friesland')

    class Table(tables.Table):
        class Meta:
            model = Region

    class PaginateByDefinedOnView(DispatchHookMixin, tables.SingleTableView):
        table_class = Table
        model = Region
        paginate_by = 2
        table_data = MEMORY_DATA

        def get_queryset(self):
            return Region.objects.all().order_by('name')

    response, view = PaginateByDefinedOnView.as_view()(build_request('/'))
    assert view.get_table().paginator.per_page == 2


@pytest.mark.django_db
def test_should_pass_kwargs_to_table_constructor():

    class PassKwargsView(SimpleView):
        table_data = []

        def get_table(self, **kwargs):
            kwargs.update({'orderable': False})
            return super(PassKwargsView, self).get_table(**kwargs)

    response, view = SimpleView.as_view()(build_request('/'))
    assert view.get_table().orderable is True

    response, view = PassKwargsView.as_view()(build_request('/'))
    assert view.get_table().orderable is False


@pytest.mark.django_db
def test_should_override_table_pagination():

    class PrefixedTable(SimpleTable):
        class Meta(SimpleTable.Meta):
            prefix = 'p_'

    class PrefixedView(SimpleView):
        table_class = PrefixedTable

    class PaginationOverrideView(PrefixedView):
        table_data = MEMORY_DATA

        def get_table_pagination(self, table):
            assert isinstance(table, tables.Table)

            per_page = self.request.GET.get('%s_override' % table.prefixed_per_page_field)
            if per_page is not None:
                return {'per_page': per_page}
            return super(PaginationOverrideView, self).get_table_pagination(table)

    response, view = PaginationOverrideView.as_view()(build_request('/?p_per_page_override=2'))
    assert view.get_table().paginator.per_page == 2


def test_singletablemixin_with_non_paginated_view():
    '''
    SingleTableMixin should not assume it is mixed with a ListView

    Github issue #326
    '''

    class Table(tables.Table):
        class Meta:
            model = Region

    class View(tables.SingleTableMixin, TemplateView):
        table_class = Table
        table_data = MEMORY_DATA

        template_name = 'dummy.html'

    View.as_view()(build_request('/'))


class TableA(tables.Table):
    class Meta:
        model = Person


class TableB(tables.Table):
    class Meta:
        model = Region
        exclude = ('id', )


@pytest.mark.django_db
def test_multiTableMixin_basic():
    Person.objects.create(first_name='Jan Pieter', last_name='W')

    Region.objects.create(name='Zuid-Holland')
    Region.objects.create(name='Noord-Holland')

    class View(tables.MultiTableMixin, TemplateView):
        tables = (TableA, TableB)
        tables_data = (Person.objects.all(), Region.objects.all())
        template_name = 'multiple.html'

    response = View.as_view()(build_request('/'))
    response.render()

    html = response.rendered_content

    assert 'table_0-sort=first_name' in html
    assert 'table_1-sort=name' in html

    assert '<td class="first_name">Jan Pieter</td>' in html
    assert '<td class="name">Zuid-Holland</td>' in html


@pytest.mark.django_db
def test_multiTableMixin_basic_alternative():
    Person.objects.create(first_name='Jan Pieter', last_name='W')

    Region.objects.create(name='Zuid-Holland')
    Region.objects.create(name='Noord-Holland')

    class View(tables.MultiTableMixin, TemplateView):
        tables = (
            TableA(Person.objects.all()),
            TableB(Region.objects.all())
        )
        template_name = 'multiple.html'

    response = View.as_view()(build_request('/'))
    response.render()

    html = response.rendered_content

    assert 'table_0-sort=first_name' in html
    assert 'table_1-sort=name' in html

    assert '<td class="first_name">Jan Pieter</td>' in html
    assert '<td class="name">Zuid-Holland</td>' in html


def test_multiTableMixin_without_tables():
    class View(tables.MultiTableMixin, TemplateView):
        template_name = 'multiple.html'

    with pytest.raises(ImproperlyConfigured):
        View.as_view()(build_request('/'))


def test_multiTableMixin_with_empty_get_tables_list():
    class View(tables.MultiTableMixin, TemplateView):
        template_name = 'multiple.html'

        def get_tables(self):
            return []

    response = View.as_view()(build_request('/'))
    response.render()

    html = response.rendered_content
    assert '<h1>Multiple tables using MultiTableMixin</h1>' in html


def test_multiTableMixin_incorrect_len():

    class View(tables.MultiTableMixin, TemplateView):
        tables = (TableA, TableB)
        tables_data = (Person.objects.all(), )
        template_name = 'multiple.html'

    with pytest.raises(ImproperlyConfigured):
        View.as_view()(build_request('/'))


@pytest.mark.django_db
def test_multiTableMixin_pagination():
    NL_PROVICES = (
        'Flevoland', 'Friesland', 'Gelderland', 'Groningen', 'Limburg',
        'Noord-Brabant', 'Noord-Holland', 'Overijssel', 'Utrecht',
        'Zeeland', 'Zuid-Holland',
    )
    for name in NL_PROVICES:
        Region.objects.create(name=name)

    class View(DispatchHookMixin, tables.MultiTableMixin, TemplateView):
        tables = (
            TableB(Region.objects.all()),
            TableB(Region.objects.all())
        )
        template_name = 'multiple.html'

        table_pagination = {
            'per_page': 5
        }

    response, view = View.as_view()(build_request('/?table_1-page=3'))

    tableA, tableB = view.get_tables()

    assert tableA.page.number == 1
    assert tableB.page.number == 3


@pytest.mark.django_db
def test_View_using_get_queryset():
    '''
    Should not raise a value-error for a View using View.get_queryset()
    (test for reverting regressing in #423)
    '''
    Person.objects.create(first_name='Anton', last_name='Sam')

    class Table(tables.Table):
        class Meta(object):
            model = Person
            fields = ('first_name', 'last_name')

    class TestView(tables.SingleTableView):
        model = Person
        table_class = Table

        def get(self, request, *args, **kwargs):
            self.get_table()
            from django.http import HttpResponse
            return HttpResponse()

        def get_queryset(self):
            '''get_queryset should be called'''
            return Person.objects.all()

    TestView.as_view()(build_request())



# coding: utf-8
from __future__ import unicode_literals

from datetime import time

from django.db import models

import django_tables2 as tables


'''
Format string: https://docs.djangoproject.com/en/stable/ref/templates/builtins/#date
'''


def test_should_handle_explicit_format():
    class TestTable(tables.Table):
        time = tables.TimeColumn(format='H:i:s')

        class Meta:
            default = '—'

    table = TestTable([{'time': time(11, 11, 11)},
                       {'time': None}])
    assert table.rows[0].get_cell('time') == "11:11:11"
    assert table.rows[1].get_cell('time') == "—"


def test_should_be_used_for_timefields():
    class TimeModel(models.Model):
        field = models.TimeField()

        class Meta:
            app_label = 'django_tables2_test'

    class Table(tables.Table):
        class Meta:
            model = TimeModel

    assert type(Table.base_columns['field']) == tables.TimeColumn


def test_value_returns_a_raw_value_without_html():
    class Table(tables.Table):
        col = tables.TimeColumn(format='H:i:s')

    table = Table([{'col': time(11, 11, 11)}])
    assert table.rows[0].get_cell_value('col') == '11:11:11'



DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

INSTALLED_APPS = [
    'tests.app',
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django_tables2',
    'haystack',
]

ROOT_URLCONF = 'tests.app.urls'

SECRET_KEY = "this is super secret"

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.request'
            ],
        }
    }
]

TIME_ZONE = "Australia/Brisbane"

USE_TZ = True

HAYSTACK_CONNECTIONS = {
    'default': {
        'ENGINE': 'haystack.backends.simple_backend.SimpleEngine',
    }
}



import re
import warnings
from contextlib import contextmanager

import lxml.etree
import lxml.html
from django.core.handlers.wsgi import WSGIRequest
from django.test.client import FakePayload
from django.utils import six


def parse(html):
    # We use html instead of etree. Because etree can't parse html entities.
    return lxml.html.fromstring(html)


class assertNumQueries(object):
    '''
    Assert the number of queries made through the django ORM in a with-block
    '''
    def __init__(self, count=1):
        self.count = count
        from django.conf import settings
        settings.DEBUG = True

    def query_count(self):
        from django.db import connection
        return len(connection.queries)

    def __enter__(self):
        self.original = self.query_count()

    def __exit__(self, exc_type, exc_value, traceback):
        expected = self.original + self.count
        count = self.query_count()
        assert expected == count, 'Expected {} queries, but got {}.'.format(
            self.count, self.query_count() - self.original
        )


def attrs(xml):
    '''
    Helper function that returns a dict of XML attributes, given an element.
    '''
    return lxml.html.fromstring(xml).attrib


@contextmanager
def warns(warning_class):
    with warnings.catch_warnings(record=True) as ws:
        warnings.simplefilter('always')
        yield ws
        assert any((issubclass(w.category, DeprecationWarning) for w in ws))


def build_request(uri='/', user=None):
    '''
    Return a fresh HTTP GET / request.

    This is essentially a heavily cutdown version of Django 1.3's
    `~django.test.client.RequestFactory`.
    '''
    path, _, querystring = uri.partition('?')
    request = WSGIRequest({
        'CONTENT_TYPE': 'text/html; charset=utf-8',
        'PATH_INFO': path,
        'QUERY_STRING': querystring,
        'REMOTE_ADDR': '127.0.0.1',
        'REQUEST_METHOD': 'GET',
        'SCRIPT_NAME': '',
        'SERVER_NAME': 'testserver',
        'SERVER_PORT': '80',
        'SERVER_PROTOCOL': 'HTTP/1.1',
        'wsgi.version': (1, 0),
        'wsgi.url_scheme': 'http',
        'wsgi.input': FakePayload(b''),
        'wsgi.errors': six.StringIO(),
        'wsgi.multiprocess': True,
        'wsgi.multithread': False,
        'wsgi.run_once': False,
    })
    if user is not None:
        request.user = user
    return request


def clean_output(s):
    '''Remove double newlines with whitespace in between and reduce the level of indentation'''
    return re.sub('\n( *\n)+', '\n', s).replace('    ', '  ')



# coding: utf-8
import django_tables2 as tables

from .utils import build_request

MEMORY_DATA = [
    {'name': 'Queensland', 'country': 'Australia', 'population': 4750500},
    {'name': 'New South Wales', 'country': 'Australia', 'population': 7565500},
    {'name': 'Victoria', 'country': 'Australia', 'population': 6000000},
    {'name': 'Tasmania', 'country': 'Australia', 'population': 517000}
]


def test_has_footer_is_False_without_footer():
    class Table(tables.Table):
        name = tables.Column()
        country = tables.Column()
        population = tables.Column()

    table = Table(MEMORY_DATA)
    assert table.has_footer() is False


def test_footer():
    class Table(tables.Table):
        name = tables.Column()
        country = tables.Column(footer='Total:')
        population = tables.Column(
            footer=lambda table: sum(x['population'] for x in table.data)
        )

    table = Table(MEMORY_DATA)
    assert table.has_footer() is True

    html = table.as_html(build_request('/'))

    assert '<td>Total:</td>' in html
    assert '<td>18833000</td>' in html


def test_footer_disable_on_table():
    '''
    Showing the footer can be disabled using show_footer argument to the Table
    constructor
    '''
    class Table(tables.Table):
        name = tables.Column()
        country = tables.Column(footer='Total:')

    table = Table(MEMORY_DATA, show_footer=False)
    assert table.has_footer() is False


def test_footer_column_method():

    class SummingColumn(tables.Column):
        def render_footer(self, bound_column, table):
            return sum(bound_column.accessor.resolve(row) for row in table.data)

    class Table(tables.Table):
        name = tables.Column()
        country = tables.Column(footer='Total:')
        population = SummingColumn()

    table = Table(MEMORY_DATA)
    html = table.as_html(build_request('/'))
    assert '<td>Total:</td>' in html
    assert '<td>18833000</td>' in html



# coding: utf-8
from __future__ import unicode_literals

import pytest
from django.template import Context, Template

import django_tables2 as tables

from ..utils import build_request


def test_should_render_in_pinned_row():
    class TestOnlyPinnedTable(tables.Table):
        foo = tables.TemplateColumn('value={{ value }}')

        def __init__(self, data):
            self.pinned = data
            revised_data = []

            super(TestOnlyPinnedTable, self).__init__(revised_data)

        def get_top_pinned_data(self):
            return self.pinned

    table = TestOnlyPinnedTable([{'foo': 'bar'}])
    for row in table.rows:
        assert row.get_cell('foo') == 'value=bar'

    template = Template('{% load django_tables2 %}{% render_table table %}')
    html = template.render(Context({'request': build_request(), 'table': table}))

    assert '<td class="foo">value=bar</td>' in html


def test_should_handle_context_on_table():
    class TestTable(tables.Table):
        col_code = tables.TemplateColumn(template_code='code:{{ record.col }}-{{ foo }}')
        col_name = tables.TemplateColumn(template_name='test_template_column.html')

    table = TestTable([{'col': 'brad'}])
    assert table.rows[0].get_cell('col_code') == 'code:brad-'
    assert table.rows[0].get_cell('col_name') == 'name:brad-empty\n'

    table.context = Context({'foo': 'author'})
    assert table.rows[0].get_cell('col_code') == 'code:brad-author'
    assert table.rows[0].get_cell('col_name') == 'name:brad-author\n'

    # new table and render using the 'render_table' template tag.
    table = TestTable([{'col': 'brad'}])
    template = Template('{% load django_tables2 %}{% render_table table %}')
    html = template.render(Context({'request': build_request(), 'table': table, 'foo': 'author'}))

    assert '<td class="col_name">name:brad-author\n</td>' in html


def test_should_support_default():
    class Table(tables.Table):
        foo = tables.TemplateColumn('default={{ default }}', default='bar')

    table = Table([{}])
    assert table.rows[0].get_cell('foo') == 'default=bar'


def test_should_support_value():
    class Table(tables.Table):
        foo = tables.TemplateColumn('value={{ value }}')

    table = Table([{'foo': 'bar'}])
    assert table.rows[0].get_cell('foo') == 'value=bar'

    template = Template('{% load django_tables2 %}{% render_table table %}')
    html = template.render(Context({'request': build_request(), 'table': table}))

    assert '<td class="foo">value=bar</td>' in html


def test_should_support_column():
    class Table(tables.Table):
        tcol = tables.TemplateColumn('column={{ column.name }}')

    table = Table([{'foo': 'bar'}])
    assert table.rows[0].get_cell('tcol') == 'column=tcol'


def test_should_raise_when_called_without_template():
    with pytest.raises(ValueError):
        class Table(tables.Table):
            col = tables.TemplateColumn()


def test_should_support_value_with_curly_braces():
    '''
    https://github.com/bradleyayers/django-tables2/issues/441
    '''
    class Table(tables.Table):
        track = tables.TemplateColumn('track: {{ value }}')

    table = Table([{'track': 'Beat it {Freestyle}'}])
    assert table.rows[0].get_cell('track') == 'track: Beat it {Freestyle}'


def test_should_strip_tags_for_value():
    class Table(tables.Table):
        track = tables.TemplateColumn('<span>{{ value }}</span>')

    table = Table([{'track': 'Space Oddity'}])

    assert list(table.as_values()) == [['Track'], ['Space Oddity']]



# coding: utf-8
import pytest
from django.core.paginator import EmptyPage, PageNotAnInteger
from fudge import Fake

from django_tables2 import Column, RequestConfig, Table

from .utils import build_request

NOTSET = object()  # unique value


@pytest.yield_fixture
def table():
    yield (Fake('Table')
           .has_attr(prefixed_page_field='page',
                     prefixed_per_page_field='per_page',
                     prefixed_order_by_field='sort'))


def test_no_querystring(table):
    request = build_request('/')
    table = table.has_attr(order_by=NOTSET).expects('paginate')
    RequestConfig(request).configure(table)
    assert table.order_by is NOTSET


def test_full_querystring(table):
    request = build_request('/?page=1&per_page=5&sort=abc')
    table = (table
             .expects('paginate').with_args(page=1, per_page=5)
             .expects('order_by').with_args('abc'))
    RequestConfig(request).configure(table)


def test_partial_querystring(table):
    request = build_request('/?page=1&sort=abc')
    table = (table
             .expects('paginate').with_args(page=1, per_page=5)
             .expects('order_by').with_args('abc'))
    RequestConfig(request, paginate={'per_page': 5}).configure(table)


def test_silent_page_not_an_integer_error(table):
    request = build_request('/')
    paginator = (Fake('Paginator')
                 .expects('page').with_args(1))
    table = (table
             .has_attr(paginator=paginator)
             .expects('paginate').with_args(page='abc')
             .raises(PageNotAnInteger))

    RequestConfig(request, paginate={'page': 'abc',
                                     'silent': True}).configure(table)


def test_silent_empty_page_error(table):
    request = build_request('/')
    paginator = (Fake('Paginator')
                 .has_attr(num_pages=987)
                 .expects('page').with_args(987))
    table = (table
             .has_attr(paginator=paginator)
             .expects('paginate').with_args(page=123)
             .raises(EmptyPage))

    RequestConfig(request, paginate={'page': 123,
                                     'silent': True}).configure(table)


def test_passing_request_to_constructor():
    '''Table constructor should call RequestConfig if a request is passed.'''

    request = build_request('/?page=1&sort=abc')

    class SimpleTable(Table):
        abc = Column()

    table = SimpleTable([
        {'abc': 'bar'},
        {'abc': 'rab'}
    ], request=request)

    assert table.columns['abc'].is_ordered



# coding: utf-8
from __future__ import unicode_literals

import json

import pytest
from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import render

import django_tables2 as tables
from django_tables2.config import RequestConfig

from ..app.models import Occupation, Person, Region
from ..test_views import DispatchHookMixin
from ..utils import build_request

# Skip if tablib is not installed (required for debian packaging)
pytest.importorskip('tablib')

try:
    from django_tables2.export.export import TableExport
    from django_tables2.export.views import ExportMixin
except ImproperlyConfigured:
    pass


NAMES = [
    ('Yildiz', 'van der Kuil'),
    ('Lindi', 'Hakvoort'),
    ('Gerardo', 'Castelein'),
]

EXPECTED_CSV = '\r\n'.join(
    ('First Name,Surname', ) + tuple(','.join(name) for name in NAMES)
) + '\r\n'

EXPECTED_JSON = list([
    {'First Name': first_name, 'Surname': last_name}
    for first_name, last_name in NAMES
])


def create_test_persons():
    for first_name, last_name in NAMES:
        Person.objects.create(first_name=first_name, last_name=last_name)


class Table(tables.Table):
    first_name = tables.Column()
    last_name = tables.Column()


class View(DispatchHookMixin, ExportMixin, tables.SingleTableView):
    table_class = Table
    table_pagination = {'per_page': 1}
    model = Person  # required for ListView
    template_name = 'django_tables2/bootstrap.html'


@pytest.mark.django_db
def test_view_should_support_csv_export():
    create_test_persons()

    response, view = View.as_view()(build_request('/?_export=csv'))
    assert response.getvalue().decode('utf8') == EXPECTED_CSV

    # should just render the normal table without the _export query
    response, view = View.as_view()(build_request('/'))
    html = response.render().rendered_content

    assert 'Yildiz' in html
    assert 'Lindy' not in html


def test_exporter_should_raise_error_for_unsupported_file_type():
    table = Table([])

    with pytest.raises(TypeError):
        TableExport(table=table, export_format='exe')


@pytest.mark.django_db
def test_view_should_support_json_export():
    create_test_persons()

    response, view = View.as_view()(build_request('/?_export=json'))
    assert json.loads(response.getvalue().decode('utf8')) == EXPECTED_JSON


@pytest.mark.django_db
def test_view_should_support_custom_trigger_param():

    class View(DispatchHookMixin, ExportMixin, tables.SingleTableView):
        table_class = Table
        export_trigger_param = 'export_to'
        model = Person  # required for ListView

    create_test_persons()

    response, view = View.as_view()(build_request('/?export_to=json'))
    assert json.loads(response.getvalue().decode('utf8')) == EXPECTED_JSON


@pytest.mark.django_db
def test_view_should_support_custom_filename():

    class View(DispatchHookMixin, ExportMixin, tables.SingleTableView):
        table_class = Table
        export_name = 'people'
        model = Person  # required for ListView

    create_test_persons()

    response, view = View.as_view()(build_request('/?_export=json'))
    assert response['Content-Disposition'] == 'attachment; filename="people.json"'


@pytest.mark.django_db
def test_function_view():
    '''
    Test the code used in the docs
    '''
    create_test_persons()

    def table_view(request):
        table = Table(Person.objects.all())
        RequestConfig(request).configure(table)

        export_format = request.GET.get('_export', None)
        if TableExport.is_valid_format(export_format):
            exporter = TableExport(export_format, table)
            return exporter.response('table.{}'.format(export_format))

        return render(request, 'django_tables2/table.html', {
            'table': table
        })

    response = table_view(build_request('/?_export=csv'))
    assert response.getvalue().decode('utf8') == EXPECTED_CSV

    # must also support the normal html table.
    response = table_view(build_request('/'))
    html = response.content.decode('utf8')

    assert 'Yildiz' in html
    assert 'Lindy' not in html


def create_test_occupations():
    richard = Person.objects.create(first_name='Richard', last_name='Queener')

    vlaanderen = Region.objects.create(name='Vlaanderen', mayor=richard)
    Occupation.objects.create(name='Timmerman', boolean=True, region=vlaanderen)
    Occupation.objects.create(name='Ecoloog', boolean=False, region=vlaanderen)


class OccupationTable(tables.Table):
    name = tables.Column()
    boolean = tables.Column()
    region = tables.Column()


class OccupationView(DispatchHookMixin, ExportMixin, tables.SingleTableView):
    table_class = OccupationTable
    table_pagination = {'per_page': 1}
    model = Occupation
    template_name = 'django_tables2/bootstrap.html'


@pytest.mark.django_db
def test_exporting_should_work_with_foreign_keys():
    create_test_occupations()

    response, view = OccupationView.as_view()(build_request('/?_export=xls'))
    data = response.content
    # binary data, so not possible to compare to an exact expectation
    assert data.find('Vlaanderen'.encode())
    assert data.find('Ecoloog'.encode())
    assert data.find('Timmerman'.encode())


@pytest.mark.django_db
def test_exporting_should_work_with_foreign_key_fields():
    create_test_occupations()

    class OccupationWithForeignKeyFieldsTable(tables.Table):
        name = tables.Column()
        boolean = tables.Column()
        region = tables.Column()
        mayor = tables.Column(accessor='region.mayor.first_name')

    class View(DispatchHookMixin, ExportMixin, tables.SingleTableView):
        table_class = OccupationWithForeignKeyFieldsTable
        table_pagination = {'per_page': 1}
        model = Occupation
        template_name = 'django_tables2/bootstrap.html'

    response, view = View.as_view()(build_request('/?_export=csv'))
    data = response.getvalue().decode('utf8')

    expected_csv = '\r\n'.join((
        'Name,Boolean,Region,First Name',
        'Timmerman,True,Vlaanderen,Richard',
        'Ecoloog,False,Vlaanderen,Richard\r\n'
    ))
    assert data == expected_csv


@pytest.mark.django_db
def test_exporting_exclude_columns():
    create_test_occupations()

    class OccupationExcludingView(DispatchHookMixin, ExportMixin, tables.SingleTableView):
        table_class = OccupationTable
        table_pagination = {'per_page': 1}
        model = Occupation
        template_name = 'django_tables2/bootstrap.html'
        exclude_columns = ('boolean', )

    response, view = OccupationExcludingView.as_view()(build_request('/?_export=csv'))
    data = response.getvalue().decode('utf8')

    assert data.splitlines()[0] == 'Name,Region'


@pytest.mark.django_db
def test_exporting_unicode_data():
    unicode_name = '木匠'
    Occupation.objects.create(name=unicode_name)

    expected_csv = 'Name,Boolean,Region\r\n{},,\r\n'.format(unicode_name)

    response, view = OccupationView.as_view()(build_request('/?_export=csv'))
    assert response.getvalue().decode('utf8') == expected_csv

    # smoke tests, hard to test this binary format for string containment
    response, view = OccupationView.as_view()(build_request('/?_export=xls'))
    data = response.content
    assert len(data) > len(expected_csv)

    response, view = OccupationView.as_view()(build_request('/?_export=xlsx'))
    data = response.content
    assert len(data) > len(expected_csv)


def test_exporting_unicode_header():
    unicode_header = 'hé'

    class Table(tables.Table):
        name = tables.Column(verbose_name=unicode_header)

    exporter = TableExport('csv', Table([]))
    response = exporter.response()
    assert response.getvalue().decode('utf8') == unicode_header + '\r\n'

    exporter = TableExport('xls', Table([]))
    # this would fail if the header contains unicode and string converstion is attempted.
    exporter.export()



# coding: utf-8
import pytest
from django.db import models
from django.utils import six

from django_tables2.utils import (Accessor, AttributeDict, OrderBy, OrderByTuple, Sequence, call_with_appropriate,
                                  computed_values, segment, signature)


def test_orderbytuple():
    obt = OrderByTuple(('a', 'b', 'c'))
    assert obt == (OrderBy('a'), OrderBy('b'), OrderBy('c'))

    # indexing
    assert obt[0] == OrderBy('a')
    assert obt['b'] == OrderBy('b')
    with pytest.raises(KeyError):
        obt['d']
    with pytest.raises(TypeError):
        obt[('tuple', )]

    # .get
    sentinel = object()
    assert obt.get('b', sentinel) is obt['b']  # keying
    assert obt.get('-', sentinel) is sentinel
    assert obt.get(0, sentinel) is obt['a']  # indexing
    assert obt.get(3, sentinel) is sentinel

    # .opposite
    assert OrderByTuple(('a', '-b', 'c')).opposite == ('-a', 'b', '-c')

    # in
    assert 'a' in obt and '-a' in obt


def test_orderbytuple_sort_key_multiple():
    obt = OrderByTuple(('a', '-b'))
    items = [
        {'a': 1, 'b': 2},
        {'a': 1, 'b': 3},
    ]
    assert sorted(items, key=obt.key) == [
        {'a': 1, 'b': 3},
        {'a': 1, 'b': 2},
    ]


def test_orderbytuple_sort_key_empty_comes_first():
    obt = OrderByTuple(('a'))
    items = [
        {'a': 1},
        {'a': ''},
        {'a': 2},
    ]
    if six.PY3:
        assert sorted(items, key=obt.key) == [
            {'a': ''},
            {'a': 1},
            {'a': 2},
        ]
    else:
        assert sorted(items, key=obt.key) == [
            {'a': 1},
            {'a': 2},
            {'a': ''},
        ]


def test_orderby():
    a = OrderBy('a')
    assert 'a' == a
    assert 'a' == a.bare
    assert '-a' == a.opposite
    assert True is a.is_ascending
    assert False is a.is_descending

    b = OrderBy('-b')
    assert '-b' == b
    assert 'b' == b.bare
    assert 'b' == b.opposite
    assert True is b.is_descending
    assert False is b.is_ascending


def test_accessor():
    x = Accessor('0')
    assert 'B' == x.resolve('Brad')

    x = Accessor('1')
    assert 'r' == x.resolve('Brad')

    x = Accessor('2.upper')
    assert 'A' == x.resolve('Brad')

    x = Accessor('2.upper.__len__')
    assert 1 == x.resolve('Brad')

    x = Accessor('')
    assert 'Brad' == x.resolve('Brad')


def test_accessor_wont_honors_alters_data():
    class Foo(object):
        deleted = False

        def delete(self):
            self.deleted = True
        delete.alters_data = True

    foo = Foo()
    with pytest.raises(ValueError):
        Accessor('delete').resolve(foo)
    assert foo.deleted is False


def test_accessor_can_be_quiet():
    foo = {}
    assert Accessor('bar').resolve(foo, quiet=True) is None


class AccessorTestModel(models.Model):
    foo = models.CharField(max_length=20)

    class Meta:
        app_label = 'django_tables2_test'


def test_accessor_can_return_field():
    context = AccessorTestModel(foo='bar')
    assert type(Accessor('foo').get_field(context)) == models.CharField


def test_accessor_returns_None_when_doesnt_exist():
    context = AccessorTestModel(foo='bar')
    assert Accessor('bar').get_field(context) is None


def test_accessor_returns_None_if_not_a_model():
    context = {'bar': 234}
    assert Accessor('bar').get_field(context) is None


def test_accessor_penultimate():
    context = {
        'a': {
            'a': 1,
            'b': {
                'c': 2,
                'd': 4
            }
        }
    }

    assert Accessor('a.b.c').penultimate(context) == (context['a']['b'], 'c')

    assert Accessor('a.b.c.d.e').penultimate(context) == (None, 'e')


def test_attribute_dict_handles_escaping():
    x = AttributeDict({'x': '"\'x&'})
    assert x.as_html() == 'x="&quot;&#39;x&amp;"'


def test_computed_values_supports_shallow_structures():
    x = computed_values({'foo': lambda: 'bar'})
    assert x == {'foo': 'bar'}


def test_computed_values_supports_nested_structures():
    x = computed_values({'foo': lambda: {'bar': lambda: 'baz'}})
    assert x == {'foo': {'bar': 'baz'}}


def test_computed_values_with_argument():
    x = computed_values({
        'foo': lambda y: {
            'bar': lambda y: 'baz-{}'.format(y)
        }
    }, kwargs=dict(y=2))
    assert x == {'foo': {'bar': 'baz-2'}}


def test_segment_should_return_all_candidates():
    assert set(segment(("a", "-b", "c"), {
        'x': 'a',
        'y': ('b', '-c'),
        '-z': ('b', '-c'),
    })) == {
        ('x', '-y'),
        ('x', 'z'),
    }


def test_sequence_multiple_ellipsis():
    sequence = Sequence(['foo', '...', 'bar', '...'])

    with pytest.raises(ValueError):
        sequence.expand(['foo'])


def test_signature():
    def foo(bar, baz):
        pass

    args, keywords = signature(foo)
    assert args == ('bar', 'baz')
    assert keywords is None


def test_signature_method():
    class Foo(object):
        def foo(self):
            pass

        def bar(self, bar, baz):
            pass

        def baz(self, bar, *bla, **boo):
            pass

    obj = Foo()
    args, keywords = signature(obj.foo)
    assert args == ()
    assert keywords is None

    args, keywords = signature(obj.bar)
    assert args == ('bar', 'baz')
    assert keywords is None

    args, keywords = signature(obj.baz)
    assert args == ('bar', )
    assert keywords == 'boo'


def test_signature_catch_all_kwargs():
    def foo(bar, baz, **kwargs):
        pass

    args, keywords = signature(foo)
    assert args == ('bar', 'baz')
    assert keywords == 'kwargs'


def test_call_with_appropriate():

    def foo():
        return 'bar'

    assert call_with_appropriate(foo, {
        'a': 'd',
        'c': 'e'
    }) == 'bar'

    def bar(baz):
        return baz

    assert call_with_appropriate(bar, dict(baz=23)) == 23



# coding: utf-8
from __future__ import unicode_literals

import pytest
from django.template import Context, Template
from django.test import TransactionTestCase
from django.utils.translation import override as translation_override
from django.utils.translation import ugettext_lazy

import django_tables2 as tables
from django_tables2.config import RequestConfig

from .app.models import Person
from .utils import build_request, parse


def test_template_override_in_settings(settings):
    settings.DJANGO_TABLES2_TEMPLATE = 'foo/bar.html'

    class Table(tables.Table):
        column = tables.Column()

    table = Table({})
    assert table.template == 'foo/bar.html'


class CountryTable(tables.Table):
    name = tables.Column()
    capital = tables.Column(orderable=False,
                            verbose_name=ugettext_lazy("Capital"))
    population = tables.Column(verbose_name='Population Size')
    currency = tables.Column(visible=False)
    tld = tables.Column(visible=False, verbose_name='Domain')
    calling_code = tables.Column(accessor='cc',
                                 verbose_name='Phone Ext.')


MEMORY_DATA = [
    {'name': 'Germany', 'capital': 'Berlin', 'population': 83,
     'currency': 'Euro (€)', 'tld': 'de', 'cc': 49},
    {'name': 'France', 'population': 64, 'currency': 'Euro (€)',
     'tld': 'fr', 'cc': 33},
    {'name': 'Netherlands', 'capital': 'Amsterdam', 'cc': '31'},
    {'name': 'Austria', 'cc': 43, 'currency': 'Euro (€)',
     'population': 8}
]


def test_as_html():
    request = build_request('/')
    table = CountryTable(MEMORY_DATA)
    root = parse(table.as_html(request))
    assert len(root.findall('.//thead/tr')) == 1
    assert len(root.findall('.//thead/tr/th')) == 4
    assert len(root.findall('.//tbody/tr')) == 4
    assert len(root.findall('.//tbody/tr/td')) == 16

    # no data with no empty_text
    table = CountryTable([])
    root = parse(table.as_html(request))
    assert 1 == len(root.findall('.//thead/tr'))
    assert 4 == len(root.findall('.//thead/tr/th'))
    assert 0 == len(root.findall('.//tbody/tr'))

    # no data WITH empty_text
    table = CountryTable([], empty_text='this table is empty')
    root = parse(table.as_html(request))
    assert 1 == len(root.findall('.//thead/tr'))
    assert 4 == len(root.findall('.//thead/tr/th'))
    assert 1 == len(root.findall('.//tbody/tr'))
    assert 1 == len(root.findall('.//tbody/tr/td'))
    assert int(root.find('.//tbody/tr/td').get('colspan')) == len(root.findall('.//thead/tr/th'))
    assert root.find('.//tbody/tr/td').text == 'this table is empty'

    # data without header
    table = CountryTable(MEMORY_DATA, show_header=False)
    root = parse(table.as_html(request))
    assert len(root.findall('.//thead')) == 0
    assert len(root.findall('.//tbody/tr')) == 4
    assert len(root.findall('.//tbody/tr/td')) == 16

    # with custom template
    table = CountryTable([], template='django_tables2/table.html')
    table.as_html(request)


def test_custom_rendering():
    '''For good measure, render some actual templates.'''
    countries = CountryTable(MEMORY_DATA)
    context = Context({'countries': countries})

    # automatic and manual column verbose names
    template = Template('{% for column in countries.columns %}{{ column }}/'
                        '{{ column.name }} {% endfor %}')
    result = ('Name/name Capital/capital Population Size/population '
              'Phone Ext./calling_code ')
    assert result == template.render(context)

    # row values
    template = Template('{% for row in countries.rows %}{% for value in row %}'
                        '{{ value }} {% endfor %}{% endfor %}')
    result = ('Germany Berlin 83 49 France — 64 33 Netherlands Amsterdam '
              '— 31 Austria — 8 43 ')
    assert result == template.render(context)


@pytest.mark.django_db
class TestQueries(TransactionTestCase):
    def test_as_html_db_queries(self):
        class PersonTable(tables.Table):
            class Meta:
                model = Person

        request = build_request('/')

        with self.assertNumQueries(1):
            PersonTable(Person.objects.all()).as_html(request)

    def test_render_table_db_queries(self):
        Person.objects.create(first_name='brad', last_name='ayers')
        Person.objects.create(first_name='davina', last_name='adisusila')

        class PersonTable(tables.Table):
            class Meta:
                model = Person
                per_page = 1
        request = build_request('/')

        with self.assertNumQueries(2):
            # one query for pagination: .count()
            # one query for page records: .all()[start:end]
            request = build_request('/')
            table = PersonTable(Person.objects.all())
            RequestConfig(request).configure(table)
            # render
            (Template('{% load django_tables2 %}{% render_table table %}')
             .render(Context({'table': table, 'request': request})))


def test_localization_check(settings):
    def get_cond_localized_table(localizeit=None):
        '''
        helper function for defining Table class conditionally
        '''
        class TestTable(tables.Table):
            name = tables.Column(verbose_name="my column", localize=localizeit)
        return TestTable

    simple_test_data = [{'name': 1234.5}]
    expected_results = {
        None: '1234.5',
        False: '1234.5',
        True: '1 234,5'  # non-breaking space
    }
    request = build_request('/')

    # no localization
    html = get_cond_localized_table(None)(simple_test_data).as_html(request)
    assert '<td class="name">{0}</td>'.format(expected_results[None]) in html

    # unlocalize
    html = get_cond_localized_table(False)(simple_test_data).as_html(request)
    assert '<td class="name">{0}</td>'.format(expected_results[False]) in html

    settings.USE_L10N = True
    settings.USE_THOUSAND_SEPARATOR = True

    with translation_override('pl'):
        # with default polish locales and enabled thousand separator
        # 1234.5 is formatted as "1 234,5" with nbsp
        html = get_cond_localized_table(True)(simple_test_data).as_html(request)
        assert '<td class="name">{0}</td>'.format(expected_results[True]) in html

        # with localize = False there should be no formatting
        html = get_cond_localized_table(False)(simple_test_data).as_html(request)
        assert '<td class="name">{0}</td>'.format(expected_results[False]) in html

        # with localize = None and USE_L10N = True
        # there should be the same formatting as with localize = True
        html = get_cond_localized_table(None)(simple_test_data).as_html(request)
        assert '<td class="name">{0}</td>'.format(expected_results[True]) in html


def test_localization_check_in_meta(settings):
    class TableNoLocalize(tables.Table):
        name = tables.Column(verbose_name='my column')

        class Meta:
            default = '---'

    class TableLocalize(tables.Table):
        name = tables.Column(verbose_name='my column')

        class Meta:
            default = '---'
            localize = ('name', )

    class TableUnlocalize(tables.Table):
        name = tables.Column(verbose_name='my column')

        class Meta:
            default = '---'
            unlocalize = ('name', )

    class TableLocalizePrecedence(tables.Table):
        name = tables.Column(verbose_name='my column')

        class Meta:
            default = '---'
            unlocalize = ('name', )
            localize = ('name', )

    simple_test_data = [{'name': 1234.5}]
    expected_results = {
        None: '1234.5',
        False: '1234.5',
        True: '1{0}234,5'.format(' ')  # non-breaking space
    }
    request = build_request('/')
    # No localize
    html = TableNoLocalize(simple_test_data).as_html(request)
    assert '<td class="name">{0}</td>'.format(expected_results[None]) in html

    settings.USE_L10N = True
    settings.USE_THOUSAND_SEPARATOR = True

    with translation_override('pl'):
        # the same as in localization_check.
        # with localization and polish locale we get formatted output
        html = TableNoLocalize(simple_test_data).as_html(request)
        assert '<td class="name">{0}</td>'.format(expected_results[True]) in html

        # localize
        html = TableLocalize(simple_test_data).as_html(request)
        assert '<td class="name">{0}</td>'.format(expected_results[True]) in html

        # unlocalize
        html = TableUnlocalize(simple_test_data).as_html(request)
        assert '<td class="name">{0}</td>'.format(expected_results[False]) in html

        # test unlocalize higher precedence
        html = TableLocalizePrecedence(simple_test_data).as_html(request)
        assert '<td class="name">{0}</td>'.format(expected_results[False]) in html


def test_localization_of_pagination_string():
    class Table(tables.Table):
        foo = tables.Column(verbose_name='my column')
        bar = tables.Column()

        class Meta:
            default = '---'

    table = Table(map(lambda x: [x, x + 100], range(40)))
    request = build_request('/')
    RequestConfig(request, paginate={'per_page': 10}).configure(table)

    with translation_override('en'):
        assert 'Page 1 of 4' in table.as_html(request)

    with translation_override('nl'):
        assert 'Pagina 1 van 4' in table.as_html(request)

    with translation_override('it'):
        assert 'Pagina 1 di 4' in table.as_html(request)

    with translation_override('nb'):
        assert 'Side 1 av 4' in table.as_html(request)


class BootstrapTable(CountryTable):
    class Meta:
        template = 'django_tables2/bootstrap.html'
        prefix = 'bootstrap-'
        per_page = 2


def test_boostrap_template():
    table = BootstrapTable(MEMORY_DATA)
    request = build_request('/')
    RequestConfig(request).configure(table)

    template = Template('{% load django_tables2 %}{% render_table table %}')
    html = template.render(Context({'request': request, 'table': table}))
    root = parse(html)
    assert len(root.findall('.//thead/tr')) == 1
    assert len(root.findall('.//thead/tr/th')) == 4
    assert len(root.findall('.//tbody/tr')) == 2
    assert len(root.findall('.//tbody/tr/td')) == 8

    assert root.find('./ul[@class="pager list-inline"]/li[@class="cardinality"]/small').text.strip() == 'Page 1 of 2'
    # make sure the link is prefixed
    assert root.find('./ul[@class="pager list-inline"]/li[@class="next"]/a').get('href') == '?bootstrap-page=2'


class SemanticTable(CountryTable):
    class Meta:
        template = 'django_tables2/semantic.html'
        prefix = 'semantic-'
        per_page = 2


def test_semantic_template():
    table = SemanticTable(MEMORY_DATA)
    request = build_request('/')
    RequestConfig(request).configure(table)

    template = Template('{% load django_tables2 %}{% render_table table %}')
    html = template.render(Context({'request': request, 'table': table}))
    root = parse(html)
    assert len(root.findall('.//thead/tr')) == 1
    assert len(root.findall('.//thead/tr/th')) == 4
    assert len(root.findall('.//tbody/tr')) == 2
    assert len(root.findall('.//tbody/tr/td')) == 8

    pager = './/tfoot/tr/th/div[@class="ui right floated pagination menu"]/div[@class="item"]'
    assert root.find(pager).text.strip() == 'Page 1 of 2'
    # make sure the link is prefixed
    next_page = './/tfoot/tr/th/div[@class="ui right floated pagination menu"]/a[@class="icon item"]'
    assert root.find(next_page).get('href') == '?semantic-page=2'


def test_bootstrap_responsive_template():
    class BootstrapResponsiveTable(BootstrapTable):
        class Meta(BootstrapTable.Meta):
            template = 'django_tables2/bootstrap-responsive.html'

    table = BootstrapResponsiveTable(MEMORY_DATA)
    request = build_request('/')
    RequestConfig(request).configure(table)

    template = Template('{% load django_tables2 %}{% render_table table %}')
    html = template.render(Context({'request': request, 'table': table}))
    root = parse(html)
    assert len(root.findall('.//thead/tr')) == 1
    assert len(root.findall('.//thead/tr/th')) == 4
    assert len(root.findall('.//tbody/tr')) == 2
    assert len(root.findall('.//tbody/tr/td')) == 8

    pager = './/ul/li[@class="cardinality"]/small'
    assert root.find(pager).text.strip() == 'Page 1 of 2'



# coding: utf-8
import django_tables2 as tables

from .models import Country, Person


class CountryTable(tables.Table):
    name = tables.Column()
    population = tables.Column()
    tz = tables.Column(verbose_name='time zone')
    visits = tables.Column()
    summary = tables.Column(order_by=('name', 'population'))

    class Meta:
        model = Country


class ThemedCountryTable(CountryTable):
    class Meta:
        attrs = {'class': 'paleblue'}


class BootstrapTable(tables.Table):

    country = tables.RelatedLinkColumn()

    class Meta:
        model = Person
        template = 'django_tables2/bootstrap.html'
        attrs = {'class': 'table table-bordered table-striped table-hover'}
        exclude = ('friendly', )


class SemanticTable(tables.Table):

    country = tables.RelatedLinkColumn()

    class Meta:
        model = Person
        template = 'django_tables2/semantic.html'
        # attrs = {'class': 'ui table table-bordered table-striped table-hover'}
        exclude = ('friendly', )


class PersonTable(tables.Table):

    class Meta:
        model = Person
        # template = 'django_tables2/bootstrap.html'



# coding: utf-8
from __future__ import unicode_literals

from os.path import dirname, join

import pytest
from django.core.files.base import ContentFile
from django.core.files.storage import FileSystemStorage
from django.db import models
from django.db.models.fields.files import FieldFile

import django_tables2 as tables

from ..utils import parse


@pytest.yield_fixture
def storage():
    '''Provide a storage that exposes the test templates'''
    root = join(dirname(__file__), '..', 'app', 'templates')
    yield FileSystemStorage(location=root, base_url='/baseurl/')


@pytest.yield_fixture
def column():
    yield tables.FileColumn(attrs={
        'span': {'class': 'span'},
        'a': {'class': 'a'}
    })


@pytest.yield_fixture
def column_with_text():
    yield tables.FileColumn(text='Download')


def test_should_be_used_for_filefields():
    class FileModel(models.Model):
        field = models.FileField()

        class Meta:
            app_label = 'django_tables2_test'

    class Table(tables.Table):
        class Meta:
            model = FileModel

    assert type(Table.base_columns['field']) == tables.FileColumn


def test_filecolumn_supports_storage_file(column, storage):
    file_ = storage.open('child/foo.html')
    try:
        root = parse(column.render(value=file_, record=None))
    finally:
        file_.close()
    path = file_.name
    assert root.tag == 'span'
    assert root.attrib == {'class': 'span exists', 'title': path}
    assert root.text == 'foo.html'


def test_filecolumn_supports_contentfile(column):
    name = 'foobar.html'
    file_ = ContentFile('')
    file_.name = name
    root = parse(column.render(value=file_, record=None))
    assert root.tag == 'span'
    assert root.attrib == {'title': name, 'class': 'span'}
    assert root.text == 'foobar.html'


def test_filecolumn_supports_fieldfile(column, storage):
    field = models.FileField(storage=storage)
    name = 'child/foo.html'
    fieldfile = FieldFile(instance=None, field=field, name=name)
    root = parse(column.render(value=fieldfile, record=None))
    assert root.tag == 'a'
    assert root.attrib == {
        'class': 'a exists',
        'title': name,
        'href': '/baseurl/child/foo.html'
    }
    assert root.text == 'foo.html'

    # Now try a file that doesn't exist
    name = 'child/does_not_exist.html'
    fieldfile = FieldFile(instance=None, field=field, name=name)
    html = column.render(value=fieldfile, record=None)
    root = parse(html)
    assert root.tag == 'a'
    assert root.attrib == {
        'class': 'a missing',
        'title': name,
        'href': '/baseurl/child/does_not_exist.html'
    }
    assert root.text == 'does_not_exist.html'


def test_filecolumn_text_custom_value(column_with_text, storage):
    name = 'foobar.html'
    file_ = ContentFile('')
    file_.name = name
    root = parse(column_with_text.render(value=file_, record=None))
    assert root.tag == 'span'
    assert root.attrib == {'title': name, 'class': ''}
    assert root.text == 'Download'




from django_tables2.data import TableListData


def generator(max_value):
    for i in range(max_value):
        yield {
            'foo': i,
            'bar': chr(i),
            'baz': hex(i),
            'inv': max_value - i
        }


def test_TableListData_basic_list():
    list_data = list(generator(100))
    data = TableListData(list_data, object())

    assert len(list_data) == len(data)
    assert data.verbose_name == 'item'
    assert data.verbose_name_plural == 'items'


def test_TableListData_with_verbose_name():
    '''
    TableListData uses the attributes on the listlike object to generate
    it's verbose_name.
    '''
    class listlike(list):
        verbose_name = 'unit'
        verbose_name_plural = 'units'

    list_data = listlike(generator(100))
    data = TableListData(list_data, object())

    assert len(list_data) == len(data)
    assert data.verbose_name == 'unit'
    assert data.verbose_name_plural == 'units'



# encoding: utf-8
import pytest
from django.db import models

import django_tables2 as tables
from django_tables2.rows import BoundRow, BoundRows

from .utils import build_request, parse


class PinnedObj(object):

    def __init__(self, name, age):
        self.name = name
        self.age = age


class SimpleTable(tables.Table):
    name = tables.Column()
    occupation = tables.Column()
    age = tables.Column()

    def get_top_pinned_data(self):
        return [
            PinnedObj("Ron", 90),
            PinnedObj("Jon", 10),
        ]

    def get_bottom_pinned_data(self):
        return [{'occupation': 'Sum age', 'age': 130}]


def test_bound_rows_with_pinned_data():
    record = {'name': 'Grzegorz', 'age': 30, 'occupation': 'programmer'}
    table = SimpleTable([record])
    row = table.rows[0]

    with pytest.raises(IndexError):
        table.rows[1]

    with pytest.raises(IndexError):
        row.get_cell(3)

    assert row.get_cell('name') == record['name']
    assert row.get_cell('occupation') == record['occupation']
    assert row.get_cell('age') == record['age']

    with pytest.raises(KeyError):
        row.get_cell('gamma')

    assert 'name' in row
    assert 'occupation' in row
    assert 'gamma' not in row


def test_as_html():
    '''
    Ensure that html render correctly.
    '''
    request = build_request('/')
    table = SimpleTable([{'name': 'Grzegorz', 'age': 30, 'occupation': 'programmer'}])
    root = parse(table.as_html(request))

    # One row for header
    assert len(root.findall('.//thead/tr')) == 1

    # In the header should be 3 cell.
    assert len(root.findall('.//thead/tr/th')) == 3

    # In the body, should be one original record and 3 pinned rows.
    assert len(root.findall('.//tbody/tr')) == 4
    assert len(root.findall('.//tbody/tr/td')) == 12

    # First top pinned row.
    tr = root.findall('.//tbody/tr')
    td = tr[0].findall('td')
    assert td[0].text == "Ron"
    assert td[1].text == table.default
    assert td[2].text == "90"

    # Second top pinned row.
    td = tr[1].findall('td')
    assert td[0].text == "Jon"
    assert td[1].text == table.default
    assert td[2].text == '10'

    # Original row
    td = tr[2].findall('td')
    assert td[0].text == "Grzegorz"
    assert td[1].text == 'programmer'
    assert td[2].text == '30'

    # First bottom pinned row.
    td = tr[3].findall('td')
    assert td[0].text == table.default
    assert td[1].text == 'Sum age'
    assert td[2].text == '130'


def test_pinned_row_attrs():
    '''
    Testing attrs for pinned rows.
    '''

    pinned_row_attrs = {
        'class': 'super-mega-row',
        'data-foo': 'bar'
    }

    request = build_request('/')
    record = {'name': 'Grzegorz', 'age': 30, 'occupation': 'programmer'}
    table = SimpleTable([record], pinned_row_attrs=pinned_row_attrs)
    html = table.as_html(request)

    assert 'pinned-row' in html
    assert 'super-mega-row' in html
    assert 'data-foo' in html


def test_ordering():
    '''
    Change sorting should not change ordering pinned rows.
    '''
    request = build_request('/')
    records = [
        {'name': 'Alex', 'age': 42, 'occupation': 'programmer'},
        {'name': 'Greg', 'age': 30, 'occupation': 'programmer'},
    ]

    table = SimpleTable(records, order_by='age')
    root = parse(table.as_html(request))
    tr = root.findall('.//tbody/tr')
    assert tr[0].findall('td')[2].text == '90'
    assert tr[1].findall('td')[2].text == '10'
    assert tr[2].findall('td')[2].text == '30'
    assert tr[3].findall('td')[2].text == '42'
    assert tr[4].findall('td')[2].text == '130'

    table = SimpleTable(records, order_by='-age')
    root = parse(table.as_html(request))
    tr = root.findall('.//tbody/tr')
    assert tr[0].findall('td')[2].text == '90'
    assert tr[1].findall('td')[2].text == '10'
    assert tr[2].findall('td')[2].text == '42'
    assert tr[3].findall('td')[2].text == '30'
    assert tr[4].findall('td')[2].text == '130'


def test_bound_rows_getitem():
    '''
    Testing BoundRows.__getitem__() method.
    Checking the return class for simple value and for slice.
    Ensure that inside of BoundRows pinned rows are included in length.
    '''
    records = [
        {'name': 'Greg', 'age': 30, 'occupation': 'policeman'},
        {'name': 'Alex', 'age': 42, 'occupation': 'programmer'},
        {'name': 'John', 'age': 72, 'occupation': 'official'},
    ]

    table = SimpleTable(records, order_by='age')
    assert isinstance(table.rows[0], BoundRow) is True
    assert isinstance(table.rows[0:2], BoundRows) is True
    assert table.rows[0:2][0].get_cell('name') == 'Greg'
    assert len(table.rows[:]) == 6


def test_uniterable_pinned_data():
    '''
    Ensure that, when data for pinned rows are not iterable,
    the ValueError exception will be raised.
    '''
    class FooTable(tables.Table):
        col = tables.Column()

        def get_top_pinned_data(self):
            return 1

    tab = FooTable([1, 2, 3])

    with pytest.raises(ValueError):
        for row in tab.rows:
            pass



# coding: utf-8
from __future__ import unicode_literals

import pytest

from django_tables2.data import TableData, TableListData, TableQuerysetData

from .app.models import Person


def test_TableData_factory_invalid_data_None():
    with pytest.raises(ValueError):
        TableData.from_data(None, table={})


def test_TableData_factory_invalid_data_int():
    with pytest.raises(ValueError):
        TableData.from_data(1, table={})


def test_TableData_factory_invalid_data_classes():
    class Klass(object):
        pass

    with pytest.raises(ValueError):
        TableData.from_data(Klass(), table={})

    class Bad(object):
        def __len__(self):
            pass

    with pytest.raises(ValueError):
        TableData.from_data(Bad(), table={})


@pytest.mark.django_db
def test_TableData_factory_valid_QuerySet():
    data = TableData.from_data(Person.objects.all(), table={})
    assert isinstance(data, TableQuerysetData)


def test_TableData_factory_valid_list_of_dicts():
    data = TableData.from_data([{'name': 'John'}, {'name': 'Pete'}], table={})
    assert isinstance(data, TableListData)
    assert len(data) == 2


def test_TableData_factory_valid_tuple_of_dicts():
    data = TableData.from_data(({'name': 'John'}, {'name': 'Pete'}), table={})
    assert isinstance(data, TableListData)
    assert len(data) == 2


def test_TableData_factory_valid_class():
    class Datasource(object):
        def __len__(self):
            return 1

        def __getitem__(self, pos):
            if pos != 0:
                raise IndexError()
            return {'a': 1}

    data = TableData.from_data(Datasource(), table={})
    assert len(data) == 1


def test_tabledata_knows_its_default_name():
    data = TableData.from_data([{}], table={})
    assert data.verbose_name == 'item'
    assert data.verbose_name_plural == 'items'


def test_tabledata_knows_its_name():
    data = TableData.from_data(Person.objects.all(), table={})

    assert data.verbose_name == 'person'
    assert data.verbose_name_plural == 'people'


# def test_tabledata_is_untouched():
#     '''
#     Ensure that data that is provided to the table (the datasource) is not
#     modified by table operations.
#     '''
#
#     MEMORY_DATA = [
#         {'i': 2, 'alpha': 'b', 'beta': 'b'},
#         {'i': 1, 'alpha': 'a', 'beta': 'c'},
#         {'i': 3, 'alpha': 'c', 'beta': 'a'},
#     ]
#
#     class Table(tables.Table):
#         i = tables.Column()
#         alpha = tables.Column()
#         beta = tables.Column()
#
#     original_data = deepcopy(MEMORY_DATA)
#
#     table = Table(MEMORY_DATA)
#     table.order_by = 'i'
#     list(table.rows)
#     assert MEMORY_DATA == original_data
#
#     table = Table(MEMORY_DATA)
#     table.order_by = 'beta'
#     list(table.rows)
#     assert MEMORY_DATA == original_data



# coding: utf-8
from __future__ import unicode_literals

import pytest
from django.core.exceptions import ImproperlyConfigured
from django.template import Context, RequestContext, Template, TemplateSyntaxError
from django.utils import six
from django.utils.six.moves.urllib.parse import parse_qs

import django_tables2 as tables
from django_tables2.config import RequestConfig

from .app.models import Person, Region
from .test_templates import MEMORY_DATA, CountryTable
from .utils import assertNumQueries, build_request, parse


def test_render_table_templatetag_invalid_type():
    template = Template('{% load django_tables2 %}{% render_table table %}')

    with pytest.raises(ValueError):
        template.render(Context({
            'request': build_request(),
            'table': dict()
        }))


def test_render_table_templatetag(settings):
    request = build_request('/')
    # ensure it works with a multi-order-by
    table = CountryTable(MEMORY_DATA, order_by=('name', 'population'))
    RequestConfig(request).configure(table)
    template = Template('{% load django_tables2 %}{% render_table table %}')
    html = template.render(Context({'request': request, 'table': table}))

    root = parse(html)
    assert len(root.findall('.//thead/tr')) == 1
    assert len(root.findall('.//thead/tr/th')) == 4
    assert len(root.findall('.//tbody/tr')) == 4
    assert len(root.findall('.//tbody/tr/td')) == 16

    # no data with no empty_text
    table = CountryTable([])
    template = Template('{% load django_tables2 %}{% render_table table %}')
    html = template.render(Context({'request': build_request('/'), 'table': table}))
    root = parse(html)
    assert len(root.findall('.//thead/tr')) == 1
    assert len(root.findall('.//thead/tr/th')) == 4
    assert len(root.findall('.//tbody/tr')) == 0

    # no data WITH empty_text
    request = build_request('/')
    table = CountryTable([], empty_text='this table is empty')
    RequestConfig(request).configure(table)
    template = Template('{% load django_tables2 %}{% render_table table %}')
    html = template.render(Context({'request': request, 'table': table}))
    root = parse(html)
    assert len(root.findall('.//thead/tr')) == 1
    assert len(root.findall('.//thead/tr/th')) == 4
    assert len(root.findall('.//tbody/tr')) == 1
    assert len(root.findall('.//tbody/tr/td')) == 1
    assert int(root.find('.//tbody/tr/td').get('colspan')) == len(root.findall('.//thead/tr/th'))
    assert root.find('.//tbody/tr/td').text == 'this table is empty'

    # variable that doesn't exist (issue #8)
    template = Template('{% load django_tables2 %}'
                        '{% render_table this_doesnt_exist %}')
    with pytest.raises(ValueError):
        settings.DEBUG = True
        template.render(Context())

    # Should still be noisy with debug off
    with pytest.raises(ValueError):
        settings.DEBUG = False
        template.render(Context())


def test_render_table_should_support_template_argument():
    table = CountryTable(MEMORY_DATA, order_by=('name', 'population'))
    template = Template('{% load django_tables2 %}'
                        '{% render_table table "dummy.html" %}')

    context = RequestContext(build_request(), {'table': table})
    assert template.render(context) == 'dummy template contents\n'


def test_render_table_template_argument_list():
    template = Template('{% load django_tables2 %}'
                        '{% render_table table template_list %}')

    context = RequestContext(build_request(), {
        'table': CountryTable(MEMORY_DATA, order_by=('name', 'population')),
        'template_list': ('dummy.html', 'child/foo.html')
    })
    assert template.render(context) == 'dummy template contents\n'


@pytest.mark.django_db
def test_render_table_supports_queryset():
    for name in ('Mackay', 'Brisbane', 'Maryborough'):
        Region.objects.create(name=name)
    template = Template('{% load django_tables2 %}{% render_table qs %}')
    html = template.render(Context({'qs': Region.objects.all(),
                                    'request': build_request('/')}))

    root = parse(html)
    assert [e.text for e in root.findall('.//thead/tr/th/a')] == ['ID', 'Name', 'Mayor']
    td = [[td.text for td in tr.findall('td')] for tr in root.findall('.//tbody/tr')]
    db = []
    for region in Region.objects.all():
        db.append([six.text_type(region.id), region.name, "—"])
    assert td == db


def test_querystring_templatetag():
    template = Template('{% load django_tables2 %}'
                        '<b>{% querystring "name"="Brad" foo.bar=value %}</b>')

    # Should be something like: <root>?name=Brad&amp;a=b&amp;c=5&amp;age=21</root>
    xml = template.render(Context({
        'request': build_request('/?a=b&name=dog&c=5'),
        'foo': {'bar': 'age'},
        'value': 21,
    }))

    # Ensure it's valid XML, retrieve the URL
    url = parse(xml).text

    qs = parse_qs(url[1:])  # everything after the ?
    assert qs['name'] == ['Brad']
    assert qs['age'] == ['21']
    assert qs['a'] == ['b']
    assert qs['c'] == ['5']


def test_querystring_templatetag_requires_request():
    template = Template('{% load django_tables2 %}{% querystring "name"="Brad" %}')
    with pytest.raises(ImproperlyConfigured):
        template.render(Context())


def test_querystring_templatetag_supports_without():
    context = Context({
        'request': build_request('/?a=b&name=dog&c=5'),
        'a_var': 'a',
    })

    template = Template('{% load django_tables2 %}'
                        '<b>{% querystring "name"="Brad" without a_var %}</b>')
    url = parse(template.render(context)).text
    qs = parse_qs(url[1:])  # trim the ?
    assert set(qs.keys()) == set(['name', 'c'])

    # Try with only exclusions
    template = Template('{% load django_tables2 %}'
                        '<b>{% querystring without "a" "name" %}</b>')
    url = parse(template.render(context)).text
    qs = parse_qs(url[1:])  # trim the ?
    assert set(qs.keys()) == set(["c"])


def test_querystring_syntax_error():
    with pytest.raises(TemplateSyntaxError):
        Template('{% load django_tables2 %}{% querystring foo= %}')


def test_title_should_only_apply_to_words_without_uppercase_letters():
    expectations = {
        'a brown fox': 'A Brown Fox',
        'a brown foX': 'A Brown foX',
        'black FBI': 'Black FBI',
        'f.b.i': 'F.B.I',
        'start 6pm': 'Start 6pm',

        # Some cyrillic samples
        'руда лисиця': 'Руда Лисиця',
        'руда лисицЯ': 'Руда лисицЯ',
        'діяльність СБУ': 'Діяльність СБУ',
        'а.б.в': 'А.Б.В',
        'вага 6кг': 'Вага 6кг',
        'у 80-их роках': 'У 80-их Роках',
    }

    for raw, expected in expectations.items():
        template = Template('{% load django_tables2 %}{{ x|title }}')
        assert template.render(Context({'x': raw})) == expected


@pytest.mark.django_db
def test_as_html_db_queries(transactional_db):
    class PersonTable(tables.Table):
        class Meta:
            model = Person

    Person.objects.create(first_name='John', last_name='Doo')

    with assertNumQueries(count=2):
        PersonTable(Person.objects.all()).as_html(build_request())



# coding: utf-8
from random import choice

from django.shortcuts import render
from django.utils.lorem_ipsum import words
from django.views.generic.base import TemplateView

from django_filters.views import FilterMixin, FilterView
from django_tables2 import MultiTableMixin, RequestConfig, SingleTableMixin, SingleTableView
from django_tables2.export.views import ExportMixin

from .data import COUNTRIES
from .filters import PersonFilter
from .models import Country, Person
from .tables import BootstrapTable, CountryTable, PersonTable, SemanticTable, ThemedCountryTable

try:
    from django.urls import reverse
except ImportError:
    # to keep backward (Django <= 1.9) compatibility
    from django.core.urlresolvers import reverse


def create_fake_data():
    # create some fake data to make sure we need to paginate
    if Country.objects.all().count() < 50:
        for country in COUNTRIES.splitlines():
            name, population = country.split(';')
            Country.objects.create(name=name, visits=0, population=int(population))

    if Person.objects.all().count() < 50:
        countries = list(Country.objects.all()) + [None]
        Person.objects.bulk_create([
            Person(name=words(3, common=False), country=choice(countries))
            for i in range(50)
        ])


def index(request):
    create_fake_data()
    table = PersonTable(Person.objects.all())
    RequestConfig(request, paginate={
        'per_page': 5
    }).configure(table)

    return render(request, 'index.html', {
        'table': table,
        'urls': (
            (reverse('tutorial'), 'Tutorial'),
            (reverse('multiple'), 'Multiple tables'),
            (reverse('filtertableview'), 'Filtered tables'),
            (reverse('singletableview'), 'Using SingleTableMixin'),
            (reverse('multitableview'), 'Using MultiTableMixin'),
            (reverse('bootstrap'), 'Using the bootstrap template'),
            (reverse('semantic'), 'Using the Semantic UI template'),
        )
    })


def multiple(request):
    qs = Country.objects.all()

    example1 = CountryTable(qs, prefix='1-')
    RequestConfig(request, paginate=False).configure(example1)

    example2 = CountryTable(qs, prefix='2-')
    RequestConfig(request, paginate={'per_page': 2}).configure(example2)

    example3 = ThemedCountryTable(qs, prefix='3-')
    RequestConfig(request, paginate={'per_page': 3}).configure(example3)

    example4 = ThemedCountryTable(qs, prefix='4-')
    RequestConfig(request, paginate={'per_page': 3}).configure(example4)

    example5 = ThemedCountryTable(qs, prefix='5-')
    example5.template = 'extended_table.html'
    RequestConfig(request, paginate={'per_page': 3}).configure(example5)

    return render(request, 'multiple.html', {
        'example1': example1,
        'example2': example2,
        'example3': example3,
        'example4': example4,
        'example5': example5,
    })


def bootstrap(request):
    '''Demonstrate the use of the bootstrap template'''

    create_fake_data()
    table = BootstrapTable(Person.objects.all(), order_by='-name')
    RequestConfig(request, paginate={'per_page': 10}).configure(table)

    return render(request, 'bootstrap_template.html', {
        'table': table
    })


def semantic(request):
    '''Demonstrate the use of the Semantic UI template'''

    create_fake_data()
    table = SemanticTable(Person.objects.all(), order_by='-name')
    RequestConfig(request, paginate={'per_page': 10}).configure(table)

    return render(request, 'semantic_template.html', {
        'table': table
    })


class ClassBased(SingleTableView):
    table_class = ThemedCountryTable
    queryset = Country.objects.all()
    template_name = 'class_based.html'


# note that this is not really the way to go because the queryset is not re
# -evaluated after the first time the view is requested.
qs = Person.objects.all()


class MultipleTables(MultiTableMixin, TemplateView):
    template_name = 'multiTable.html'
    tables = [
        PersonTable(qs),
        PersonTable(qs, exclude=('country', ))
    ]

    table_pagination = {
        'per_page': 10
    }


def tutorial(request):
    return render(request, 'tutorial.html', {'people': Person.objects.all()})


class FilteredPersonListView(SingleTableMixin, FilterView):
    table_class = PersonTable
    model = Person
    template_name = 'bootstrap_template.html'

    filterset_class = PersonFilter



# coding: utf-8
from __future__ import unicode_literals

import pytest
from django.db import models
from django.utils.safestring import SafeData, mark_safe
from django.utils.translation import ugettext_lazy

import django_tables2 as tables

from ..app.models import Person
from ..utils import build_request, parse

request = build_request('/')


def test_column_render_supports_kwargs():
    class TestColumn(tables.Column):
        def render(self, **kwargs):
            expected = {'record', 'value', 'column', 'bound_column', 'bound_row', 'table'}
            actual = set(kwargs.keys())
            assert actual == expected
            return 'success'

    class TestTable(tables.Table):
        foo = TestColumn()

    table = TestTable([{'foo': 'bar'}])
    assert table.rows[0].get_cell('foo') == 'success'


def test_column_header_should_use_titlised_verbose_name_unless_given_explicitly():
    class SimpleTable(tables.Table):
        basic = tables.Column()
        acronym = tables.Column(verbose_name='has FBI help')

    table = SimpleTable([])
    assert table.columns['basic'].header == 'Basic'
    assert table.columns['acronym'].header == 'has FBI help'


def test_should_support_safe_verbose_name():
    class SimpleTable(tables.Table):
        safe = tables.Column(verbose_name=mark_safe('<b>Safe</b>'))

    table = SimpleTable([])
    assert isinstance(table.columns['safe'].header, SafeData)


def test_should_raise_on_invalid_accessor():
    with pytest.raises(TypeError):
        class SimpleTable(tables.Table):
            column = tables.Column(accessor={})


def test_column_with_callable_accessor_should_not_have_default():
    with pytest.raises(TypeError):
        class SimpleTable(tables.Table):
            column = tables.Column(accessor=lambda: 'foo', default='')


def test_should_support_safe_verbose_name_via_model():
    class PersonTable(tables.Table):
        safe = tables.Column()

    table = PersonTable(Person.objects.all())
    assert isinstance(table.columns['safe'].header, SafeData)


def test_should_support_empty_string_as_explicit_verbose_name():
    class SimpleTable(tables.Table):
        acronym = tables.Column(verbose_name='')

    table = SimpleTable([])
    assert table.columns['acronym'].header == ''


@pytest.mark.django_db
def test_handle_verbose_name_of_many2onerel():

    class Table(tables.Table):
        count = tables.Column(accessor='info_list.count')

    Person.objects.create(first_name='bradley', last_name='ayers')
    table = Table(Person.objects.all())
    assert table.columns['count'].verbose_name == 'Information'


def test_orderable():
    class SimpleTable(tables.Table):
        name = tables.Column()

    table = SimpleTable([])
    assert table.columns['name'].orderable is True

    class SimpleTable(tables.Table):
        name = tables.Column()

        class Meta:
            orderable = False
    table = SimpleTable([])
    assert table.columns['name'].orderable is False

    class SimpleTable(tables.Table):
        name = tables.Column()

        class Meta:
            orderable = True

    table = SimpleTable([])
    assert table.columns['name'].orderable is True


def test_order_by_defaults_to_accessor():
    class SimpleTable(tables.Table):
        foo = tables.Column(accessor='bar')

    table = SimpleTable([])
    assert table.columns['foo'].order_by == ('bar', )


def test_supports_order_by():
    class SimpleTable(tables.Table):
        name = tables.Column(order_by=('last_name', '-first_name'))
        age = tables.Column()

    table = SimpleTable([], order_by=('-age', ))
    # alias
    assert table.columns['name'].order_by_alias == 'name'
    assert table.columns['age'].order_by_alias == '-age'
    # order by
    assert table.columns['name'].order_by == ('last_name', '-first_name')
    assert table.columns['age'].order_by == ('-age', )

    # now try with name ordered
    table = SimpleTable([], order_by=('-name', ))
    # alias
    assert table.columns['name'].order_by_alias == '-name'
    assert table.columns['age'].order_by_alias == 'age'
    # alias next
    assert table.columns['name'].order_by_alias.next == 'name'
    assert table.columns['age'].order_by_alias.next == 'age'
    # order by
    assert table.columns['name'].order_by == ('-last_name', 'first_name')
    assert table.columns['age'].order_by == ('age', )


def test_supports_is_ordered():
    class SimpleTable(tables.Table):
        name = tables.Column()

    # sorted
    table = SimpleTable([], order_by='name')
    assert table.columns['name'].is_ordered
    # unsorted
    table = SimpleTable([])
    assert not table.columns['name'].is_ordered


def test_translation():
    '''
    Tests different types of values for the ``verbose_name`` property of a
    column.
    '''
    class TranslationTable(tables.Table):
        text = tables.Column(verbose_name=ugettext_lazy('Text'))

    table = TranslationTable([])
    assert 'Text' == table.columns['text'].header


def test_sequence():
    '''
    Ensures that the sequence of columns is configurable.
    '''
    class TestTable(tables.Table):
        a = tables.Column()
        b = tables.Column()
        c = tables.Column()
    assert ['a', 'b', 'c'] == TestTable([]).columns.names()
    assert ['b', 'a', 'c'] == TestTable([], sequence=('b', 'a', 'c')).columns.names()

    class TestTable2(TestTable):
        class Meta:
            sequence = ('b', 'a', 'c')
    assert ['b', 'a', 'c'] == TestTable2([]).columns.names()
    assert ['a', 'b', 'c'] == TestTable2([], sequence=('a', 'b', 'c')).columns.names()

    class TestTable3(TestTable):
        class Meta:
            sequence = ('c', )
    assert ['c', 'a', 'b'] == TestTable3([]).columns.names()
    assert ['c', 'a', 'b'] == TestTable([], sequence=('c', )).columns.names()

    class TestTable4(TestTable):
        class Meta:
            sequence = ('...', )
    assert ['a', 'b', 'c'] == TestTable4([]).columns.names()
    assert ['a', 'b', 'c'] == TestTable([], sequence=('...', )).columns.names()

    class TestTable5(TestTable):
        class Meta:
            sequence = ('b', '...')
    assert ['b', 'a', 'c'] == TestTable5([]).columns.names()
    assert ['b', 'a', 'c'] == TestTable([], sequence=('b', '...')).columns.names()

    class TestTable6(TestTable):
        class Meta:
            sequence = ('...', 'b')
    assert ['a', 'c', 'b'] == TestTable6([]).columns.names()
    assert ['a', 'c', 'b'] == TestTable([], sequence=('...', 'b')).columns.names()

    class TestTable7(TestTable):
        class Meta:
            sequence = ('b', '...', 'a')
    assert ['b', 'c', 'a'] == TestTable7([]).columns.names()
    assert ['b', 'c', 'a'] == TestTable([], sequence=('b', '...', 'a')).columns.names()

    # Let's test inheritence
    class TestTable8(TestTable):
        d = tables.Column()
        e = tables.Column()
        f = tables.Column()

        class Meta:
            sequence = ('d', '...')

    class TestTable9(TestTable):
        d = tables.Column()
        e = tables.Column()
        f = tables.Column()

    assert ['d', 'a', 'b', 'c', 'e', 'f'] == TestTable8([]).columns.names()
    assert ['d', 'a', 'b', 'c', 'e', 'f'] == TestTable9([], sequence=('d', '...')).columns.names()


def test_should_support_both_meta_sequence_and_constructor_exclude():
    '''
    Issue #32 describes a problem when both ``Meta.sequence`` and
    ``Table(..., exclude=...)`` are used on a single table. The bug caused an
    exception to be raised when the table was iterated.
    '''
    class SequencedTable(tables.Table):
        a = tables.Column()
        b = tables.Column()
        c = tables.Column()

        class Meta:
            sequence = ('a', '...')

    table = SequencedTable([], exclude=('c', ))
    table.as_html(request)


def test_bound_columns_should_support_indexing():
    class SimpleTable(tables.Table):
        a = tables.Column()
        b = tables.Column()

    table = SimpleTable([])
    assert 'b' == table.columns[1].name
    assert 'b' == table.columns['b'].name


def test_cell_attrs_applies_to_td_and_th():
    class SimpleTable(tables.Table):
        a = tables.Column(attrs={'cell': {'key': 'value'}})

    # providing data ensures 1 row is rendered
    table = SimpleTable([{'a': 'value'}])
    root = parse(table.as_html(request))

    assert root.findall('.//thead/tr/th')[0].attrib == {'key': 'value', 'class': 'a orderable'}
    assert root.findall('.//tbody/tr/td')[0].attrib == {'key': 'value', 'class': 'a'}


def test_cells_are_automatically_given_column_name_as_class():
    class SimpleTable(tables.Table):
        a = tables.Column()

    table = SimpleTable([{'a': 'value'}])
    root = parse(table.as_html(request))
    assert root.findall('.//thead/tr/th')[0].attrib == {'class': 'a orderable'}
    assert root.findall('.//tbody/tr/td')[0].attrib == {'class': 'a'}


def test_th_are_given_orderable_class_if_column_is_orderable():
    class SimpleTable(tables.Table):
        a = tables.Column()
        b = tables.Column(orderable=False)

    table = SimpleTable([{'a': 'value'}])
    root = parse(table.as_html(request))
    # return classes of an element as a set
    classes = lambda x: set(x.attrib['class'].split())
    assert 'orderable' in classes(root.findall('.//thead/tr/th')[0])
    assert 'orderable' not in classes(root.findall('.//thead/tr/th')[1])

    # Now try with an ordered table
    table = SimpleTable([], order_by='a')
    root = parse(table.as_html(request))
    # return classes of an element as a set
    assert 'orderable' in classes(root.findall('.//thead/tr/th')[0])
    assert 'asc' in classes(root.findall('.//thead/tr/th')[0])
    assert 'orderable' not in classes(root.findall('.//thead/tr/th')[1])


def test_empty_values_triggers_default():
    class Table(tables.Table):
        a = tables.Column(empty_values=(1, 2), default='--')

    table = Table([{'a': 1}, {'a': 2}, {'a': 3}, {'a': 4}])
    assert [row.get_cell('a') for row in table.rows] == ['--', '--', 3, 4]


def test_register_skips_non_columns():
    from django_tables2.columns.base import library

    @library.register
    class Klass(object):
        pass

    class Table(tables.Table):
        class Meta:
            model = Person

    Table([])


def test_raises_when_using_non_supported_index():
    class Table(tables.Table):
        column = tables.Column()

    table = Table([{'column': 'foo'}])

    row = table.rows[0]
    with pytest.raises(TypeError):
        row[table]


class MyModel(models.Model):
    item1 = models.CharField(max_length=10)

    class Meta:
        app_label = 'django_tables2_tests'


class MyTable(tables.Table):
    item1 = tables.Column(verbose_name='Nice column name')

    class Meta:
        model = MyModel
        fields = ('item1', )


def test_column_params_should_be_preserved_under_inheritance():
    '''
    Github issue #337

    Columns explicitly defined on MyTable get overridden by columns implicitly
    defined on it's child.
    If the column is not redefined, the explicit definition of MyTable is used,
    preserving the specialized verbose_name defined on it.
    '''

    class MyTableA(MyTable):
        '''
        having an empty `class Meta` should not undo the explicit definition
        of column item1 in MyTable.
        '''
        class Meta(MyTable.Meta):
            pass

    class MyTableB(MyTable):
        '''
        having a non-empty `class Meta` should not undo the explicit definition
        of column item1 in MyTable.
        '''
        class Meta(MyTable.Meta):
            per_page = 22

    table = MyTable(MyModel.objects.all())
    tableA = MyTableA(MyModel.objects.all())
    tableB = MyTableB(MyModel.objects.all())

    assert table.columns['item1'].verbose_name == 'Nice column name'
    assert tableA.columns['item1'].verbose_name == 'Nice column name'
    assert tableB.columns['item1'].verbose_name == 'Nice column name'


def test_explicit_column_can_be_overridden_by_other_explicit_column():
    class MyTableC(MyTable):
        '''
        If we define a new explict item1 column, that one should be used.
        '''
        item1 = tables.Column(verbose_name='New nice column name')

    table = MyTable(MyModel.objects.all())
    tableC = MyTableC(MyModel.objects.all())

    assert table.columns['item1'].verbose_name == 'Nice column name'
    assert tableC.columns['item1'].verbose_name == 'New nice column name'


def test_override_column_class_names():
    '''
    We control the output of CSS class names for a column by overriding
    get_column_class_names
    '''
    class MyTable(tables.Table):
        population = tables.Column(verbose_name='Population')

        def get_column_class_names(self, classes_set, bound_column):
            classes_set.add('prefix-%s' % bound_column.name)
            return classes_set

    TEST_DATA = [
        {'name': 'Belgium', 'population': 11200000},
        {'name': 'Luxembourgh', 'population': 540000},
        {'name': 'France', 'population': 66000000},
    ]

    html = MyTable(TEST_DATA).as_html(build_request())

    assert '<td class="prefix-population">11200000</td>' in html


@pytest.mark.django_db
def test_computable_td_attrs():
    '''Computable attrs for columns, using table argument'''
    Person.objects.create(first_name='Jan', last_name='Pietersz.')
    Person.objects.create(first_name='Sjon', last_name='Jansen')

    class Table(tables.Table):
        person = tables.Column(attrs={
            'cell': {
                'data-length': lambda table: len(table.data)
            }
        })

    table = Table(Person.objects.all())
    html = table.as_html(request)
    assert 'data-length="2"' in html



# coding: utf-8
from __future__ import unicode_literals

from django.db import models

import django_tables2 as tables

MEMORY_DATA = [
    {'url': 'http://example.com', 'name': 'Example'},
    {'url': 'https://example.com', 'name': 'Example (https)'},
    {'url': 'ftp://example.com', 'name': 'Example (ftp)'},
]


def test_should_turn_url_into_hyperlink():
    class TestTable(tables.Table):
        url = tables.URLColumn()

    table = TestTable(MEMORY_DATA)

    assert table.rows[0].get_cell('url') == '<a href="http://example.com">http://example.com</a>'
    assert table.rows[1].get_cell('url') == '<a href="https://example.com">https://example.com</a>'


def test_should_be_used_for_urlfields():
    class URLModel(models.Model):
        field = models.URLField()

        class Meta:
            app_label = 'django_tables2_test'

    class Table(tables.Table):
        class Meta:
            model = URLModel

    assert type(Table.base_columns['field']) == tables.URLColumn


def test_text_can_be_overridden():
    class Table(tables.Table):
        url = tables.URLColumn(text='link')

    table = Table(MEMORY_DATA)

    assert table.rows[0].get_cell('url') == '<a href="http://example.com">link</a>'


def test_text_can_be_overridden_with_callable():
    class Table(tables.Table):
        url = tables.URLColumn(text=lambda record: record['name'])

    table = Table(MEMORY_DATA)

    assert table.rows[0].get_cell('url') == '<a href="http://example.com">Example</a>'
    assert table.rows[1].get_cell('url') == '<a href="https://example.com">Example (https)</a>'


def test_value_returns_a_raw_value_without_html():
    class TestTable(tables.Table):
        col = tables.URLColumn()

    table = TestTable([{'col': 'http://example.com'}])

    assert table.rows[0].get_cell_value('col') == 'http://example.com'



# coding: utf-8
from __future__ import unicode_literals

from random import randint, sample

import pytest
from django.utils.html import format_html, mark_safe, strip_tags

import django_tables2 as tables
from tests.app.models import Person

pytestmark = pytest.mark.django_db

FAKE_NAMES = (
    ('Kyle', 'Strader'),
    ('Francis', 'Fisher'),
    ('James', 'Jury'),
    ('Florentina', 'Floyd'),
    ('Mark', 'Boyd'),
    ('Simone', 'Fong'),
)


def create_Persons():
    for first, last in FAKE_NAMES:
        Person.objects.create(first_name=first, last_name=last)

    persons = list(Person.objects.all())

    # give everyone 1 to 3 friends
    for person in persons:
        person.friends.add(*sample(persons, randint(1, 3)))
        person.save()


def test_ManyToManyColumn_from_model():
    '''
    Automaticcally uses the ManyToManyColumn for a ManyToManyField, and calls the
    Models's `__str__` method to transform the model instace to string.
    '''
    create_Persons()

    class Table(tables.Table):
        name = tables.Column(accessor='name', order_by=('last_name', 'first_name'))

        class Meta:
            model = Person
            fields = ('name', 'friends')

    table = Table(Person.objects.all())

    for row in table.rows:
        friends = row.get_cell('friends').split(', ')

        for friend in friends:
            assert Person.objects.filter(first_name=friend).exists()


def test_custom_separator():
    create_Persons()

    def assert_sep(sep):
        class Table(tables.Table):
            friends = tables.ManyToManyColumn(separator=sep)

        table = Table(Person.objects.all().order_by('last_name'))
        for row in table.rows:
            friends = row.get_cell('friends').split(sep)

            for friend in friends:
                assert Person.objects.filter(first_name=friend).exists()

    # normal string, will not be escaped
    assert_sep('|')

    # html tag, would normally be escaped, but should not be escaped because
    # it is mark_safe()'ed
    assert_sep(mark_safe('<br />'))


def test_transform_returns_html():
    create_Persons()

    class Table(tables.Table):
        friends = tables.ManyToManyColumn(transform=lambda m: format_html('<span>{}</span>', m.first_name))

    table = Table(Person.objects.all().order_by('last_name'))
    for row in table.rows:
        friends = row.get_cell('friends').split(', ')
        for friend in friends:
            stripped = strip_tags(friend)
            assert Person.objects.filter(first_name=stripped).exists()


def test_orderable_is_false():
    class Table(tables.Table):
        friends = tables.ManyToManyColumn(orderable=False)

    Table([])


def test_ManyToManyColumn_complete_example():
    create_Persons()

    # add a friendless person
    remi = Person.objects.create(first_name='Remi', last_name='Barberin')

    class Table(tables.Table):
        name = tables.Column(accessor='name', order_by=('last_name', 'first_name'))
        friends = tables.ManyToManyColumn(
            transform=lambda o: o.name,
            filter=lambda o: o.order_by('-last_name')
        )

    table = Table(Person.objects.all().order_by('last_name'))
    for row in table.rows:
        friends = row.get_cell('friends')
        if friends == '-':
            assert row.get_cell('name') == remi.name
            continue

        friends = list(map(lambda o: o.split(' '), friends.split(', ')))

        assert friends == sorted(friends, key=lambda item: item[1], reverse=True)



# coding: utf-8
from __future__ import unicode_literals

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import six
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext, ugettext_lazy

from haystack import indexes

try:
    from django.urls import reverse
except ImportError:
    # to keep backward (Django <= 1.9) compatibility
    from django.core.urlresolvers import reverse


@six.python_2_unicode_compatible
class Person(models.Model):
    first_name = models.CharField(max_length=200)

    last_name = models.CharField(max_length=200, verbose_name='surname')

    occupation = models.ForeignKey(
        'Occupation',
        related_name='people',
        null=True,
        verbose_name='occupation of the person',
        on_delete=models.CASCADE
    )

    trans_test = models.CharField(
        max_length=200, blank=True,
        verbose_name=ugettext("translation test"))

    trans_test_lazy = models.CharField(
        max_length=200, blank=True,
        verbose_name=ugettext_lazy("translation test lazy"))

    safe = models.CharField(
        max_length=200, blank=True, verbose_name=mark_safe("<b>Safe</b>"))

    website = models.URLField(
        max_length=200, null=True, blank=True,
        verbose_name="web site")

    birthdate = models.DateField(null=True)

    content_type = models.ForeignKey(
        ContentType,
        null=True,
        blank=True,
        on_delete=models.CASCADE
    )
    object_id = models.PositiveIntegerField(null=True, blank=True)
    foreign_key = GenericForeignKey()

    friends = models.ManyToManyField('Person')

    class Meta:
        verbose_name = "person"
        verbose_name_plural = "people"

    def __str__(self):
        return self.first_name

    @property
    def name(self):
        return "%s %s" % (self.first_name, self.last_name)

    def get_absolute_url(self):
        return reverse('person', args=(self.pk, ))


class PersonProxy(Person):
    class Meta:
        proxy = True
        ordering = ('last_name', )


@six.python_2_unicode_compatible
class Occupation(models.Model):
    name = models.CharField(max_length=200)
    region = models.ForeignKey('Region', null=True, on_delete=models.CASCADE)
    boolean = models.BooleanField(null=True)
    boolean_with_choices = models.BooleanField(null=True, choices=(
        (True, 'Yes'),
        (False, 'No')
    ))

    def get_absolute_url(self):
        return reverse('occupation', args=(self.pk, ))

    def __str__(self):
        return self.name


@six.python_2_unicode_compatible
class Region(models.Model):
    name = models.CharField(max_length=200)
    mayor = models.OneToOneField(Person, null=True, on_delete=models.CASCADE)

    def __str__(self):
        return self.name


class PersonInformation(models.Model):
    person = models.ForeignKey(
        Person,
        related_name='info_list',
        verbose_name='Information',
        on_delete=models.CASCADE
    )


# -- haystack -----------------------------------------------------------------

class PersonIndex(indexes.SearchIndex, indexes.Indexable):
    first_name = indexes.CharField(document=True)

    def get_model(self):
        return Person

    def index_queryset(self, using=None):
        return self.get_model().objects.all()



import itertools

import django_tables2 as tables

from .utils import build_request

TEST_DATA = [
    {'name': 'Belgium', 'population': 11200000},
    {'name': 'Luxembourgh', 'population': 540000},
    {'name': 'France', 'population': 66000000},
]


def _test_counter(Table, expected='<td class="counter">0</td>'):
    table = Table(TEST_DATA)
    html = table.as_html(build_request())
    assert expected in html

    # the counter should start at zero the second time too
    table = Table(TEST_DATA)
    html = table.as_html(build_request())
    assert expected in html

    return html


def test_row_counter_using_templateColumn():
    class CountryTable(tables.Table):
        counter = tables.TemplateColumn('{{ row_counter }}')
        name = tables.Column()

    _test_counter(CountryTable)


def test_row_footer_total():
    class CountryTable(tables.Table):
        name = tables.Column()
        population = tables.Column(
            footer=lambda table: 'Total: {}'.format(
                sum(x['population'] for x in table.data)
            )
        )

    table = CountryTable(TEST_DATA)
    html = table.as_html(build_request())

    assert '<td>Total: 77740000</td>' in html



# coding: utf-8
'''Test the core table functionality.'''
from __future__ import absolute_import, unicode_literals

import copy
import itertools

import pytest
from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator

import django_tables2 as tables
from django_tables2.tables import DeclarativeColumnsMetaclass

from .utils import build_request, parse

request = build_request('/')

MEMORY_DATA = [
    {'i': 2, 'alpha': 'b', 'beta': 'b'},
    {'i': 1, 'alpha': 'a', 'beta': 'c'},
    {'i': 3, 'alpha': 'c', 'beta': 'a'},
]


class UnorderedTable(tables.Table):
    i = tables.Column()
    alpha = tables.Column()
    beta = tables.Column()


class OrderedTable(UnorderedTable):
    class Meta:
        order_by = 'alpha'


def test_omitting_data():
    with pytest.raises(TypeError):
        UnorderedTable()


def test_column_named_items():
    '''
    A column named items must not make the table fail
    https://github.com/bradleyayers/django-tables2/issues/316
    '''
    class ItemsTable(tables.Table):
        items = tables.Column()

    table = ItemsTable([{'items': 123}, {'items': 2345}])

    html = table.as_html(request)
    assert '123' in html
    assert '2345' in html


def test_declarations():
    '''Test defining tables by declaration.'''
    class GeoAreaTable(tables.Table):
        name = tables.Column()
        population = tables.Column()

    assert len(GeoAreaTable.base_columns) == 2
    assert 'name' in GeoAreaTable.base_columns
    assert not hasattr(GeoAreaTable, 'name')

    class CountryTable(GeoAreaTable):
        capital = tables.Column()

    assert len(CountryTable.base_columns) == 3
    assert 'capital' in CountryTable.base_columns

    # multiple inheritance
    class AddedMixin(tables.Table):
        added = tables.Column()

    class CityTable(GeoAreaTable, AddedMixin):
        mayor = tables.Column()

    assert len(CityTable.base_columns) == 4
    assert 'added' in CityTable.base_columns

    # overwrite a column with a non-column
    class MayorlessCityTable(CityTable):
        mayor = None

    assert len(MayorlessCityTable.base_columns) == 3


def test_metaclass_inheritance():
    class Tweaker(type):
        '''Adds an attribute "tweaked" to all classes'''
        def __new__(cls, name, bases, attrs):
            attrs['tweaked'] = True
            return super(Tweaker, cls).__new__(cls, name, bases, attrs)

    class Meta(Tweaker, DeclarativeColumnsMetaclass):
        pass

    class TweakedTableBase(tables.Table):
        __metaclass__ = Meta
        name = tables.Column()

    # Python 2/3 compatible way to enable the metaclass
    TweakedTable = Meta(str('TweakedTable'), (TweakedTableBase, ), {})

    table = TweakedTable([])
    assert 'name' in table.columns
    assert table.tweaked

    # now flip the order
    class FlippedMeta(DeclarativeColumnsMetaclass, Tweaker):
        pass

    class FlippedTweakedTableBase(tables.Table):
        name = tables.Column()

    # Python 2/3 compatible way to enable the metaclass
    FlippedTweakedTable = FlippedMeta(str('FlippedTweakedTable'), (FlippedTweakedTableBase, ), {})

    table = FlippedTweakedTable([])
    assert 'name' in table.columns
    assert table.tweaked


def test_attrs():
    class TestTable(tables.Table):
        class Meta:
            attrs = {}
    assert {} == TestTable([]).attrs

    class TestTable2(tables.Table):
        class Meta:
            attrs = {'a': 'b'}
    assert {'a': 'b'} == TestTable2([]).attrs

    class TestTable3(tables.Table):
        pass
    assert {} == TestTable3([]).attrs
    assert {'a': 'b'} == TestTable3([], attrs={'a': 'b'}).attrs

    class TestTable4(tables.Table):
        class Meta:
            attrs = {'a': 'b'}
    assert {'c': 'd'} == TestTable4([], attrs={'c': 'd'}).attrs


def test_attrs_support_computed_values():
    counter = itertools.count()

    class TestTable(tables.Table):
        class Meta:
            attrs = {'id': lambda: 'test_table_%d' % next(counter)}

    assert 'id="test_table_0"' == TestTable([]).attrs.as_html()
    assert 'id="test_table_1"' == TestTable([]).attrs.as_html()


def test_attrs_from_settings(settings):
    settings.DJANGO_TABLES2_TABLE_ATTRS = {
        'class': 'table-compact'
    }

    class Table(tables.Table):
        column = tables.Column()

    table = Table({})
    assert table.attrs == {'class': 'table-compact'}


def test_datasource_untouched():
    '''
    Ensure that data that is provided to the table (the datasource) is not
    modified by table operations.
    '''
    original_data = copy.deepcopy(MEMORY_DATA)

    table = UnorderedTable(MEMORY_DATA)
    table.order_by = 'i'
    list(table.rows)
    assert MEMORY_DATA == original_data

    table = UnorderedTable(MEMORY_DATA)
    table.order_by = 'beta'
    list(table.rows)
    assert MEMORY_DATA == original_data


def test_should_support_tuple_data_source():
    class SimpleTable(tables.Table):
        name = tables.Column()

    table = SimpleTable((
        {'name': 'brad'},
        {'name': 'davina'},
    ))

    assert len(table.rows) == 2


# @pytest.mark.django_db
# def test_should_support_haystack_data_source():
#     Person.objects.create(first_name='Foo', last_name='Bar')
#     Person.objects.create(first_name='Brad', last_name='Pitt')
#
#     from haystack.query import SearchQuerySet
#     from haystack.management.commands import update_index
#
#     update_index.Command().handle(interactive=False)
#
#     class PersonTable(tables.Table):
#         first_name = tables.Column()
#
#     table = PersonTable(SearchQuerySet().all())
#     html = table.as_html(request)
#
#     # TODO: assert that a person is actually in the produced html.
#     assert 'Brad' in html
#

def test_column_count():
    class SimpleTable(tables.Table):
        visible = tables.Column(visible=True)
        hidden = tables.Column(visible=False)

    # The columns container supports the len() builtin
    assert len(SimpleTable([]).columns) == 1


def test_column_accessor():
    class SimpleTable(UnorderedTable):
        col1 = tables.Column(accessor='alpha.upper.isupper')
        col2 = tables.Column(accessor='alpha.upper')
    table = SimpleTable(MEMORY_DATA)

    assert table.rows[0].get_cell('col1') is True
    assert table.rows[0].get_cell('col2') == 'B'


def test_exclude_columns():
    '''
    Defining ``Table.Meta.exclude`` or providing an ``exclude`` argument when
    instantiating a table should have the same effect -- exclude those columns
    from the table. It should have the same effect as not defining the
    columns originally.
    '''
    table = UnorderedTable([], exclude=('i'))
    assert table.columns.names() == ['alpha', 'beta']

    # Table.Meta: exclude=...
    class PartialTable(UnorderedTable):
        class Meta:
            exclude = ('alpha', )
    table = PartialTable([])
    assert table.columns.names() == ['i', 'beta']

    # Inheritence -- exclude in parent, add in child
    class AddonTable(PartialTable):
        added = tables.Column()
    table = AddonTable([])
    assert table.columns.names() == ['i', 'beta', 'added']

    # Inheritence -- exclude in child
    class ExcludeTable(UnorderedTable):
        added = tables.Column()

        class Meta:
            exclude = ('beta', )

    table = ExcludeTable([])
    assert table.columns.names() == ['i', 'alpha', 'added']


def test_table_exclude_property_should_override_constructor_argument():
    class SimpleTable(tables.Table):
        a = tables.Column()
        b = tables.Column()

    table = SimpleTable([], exclude=('b', ))
    assert table.columns.names() == ['a']
    table.exclude = ('a', )
    assert table.columns.names() == ['b']


def test_exclude_should_work_on_sequence_too():
    '''
    It should be possible to define a sequence on a table
    and exclude it in a child of that table.
    '''
    class PersonTable(tables.Table):
        first_name = tables.Column()
        last_name = tables.Column()
        occupation = tables.Column()

        class Meta:
            sequence = ('first_name', 'last_name', 'occupation')

    class AnotherPersonTable(PersonTable):
        class Meta(PersonTable.Meta):
            exclude = ('first_name', 'last_name')

    tableA = PersonTable([])
    assert tableA.columns.names() == ['first_name', 'last_name', 'occupation']

    tableB = AnotherPersonTable([])
    assert tableB.columns.names() == ['occupation']

    tableC = PersonTable([], exclude=('first_name'))
    assert tableC.columns.names() == ['last_name', 'occupation']


def test_pagination():
    class BookTable(tables.Table):
        name = tables.Column()

    # create some sample data
    data = []
    for i in range(100):
        data.append({'name': 'Book No. %d' % i})
    books = BookTable(data)

    # external paginator
    paginator = Paginator(books.rows, 10)
    assert paginator.num_pages == 10
    page = paginator.page(1)
    assert page.has_previous() is False
    assert page.has_next() is True

    # integrated paginator
    books.paginate(page=1)
    assert hasattr(books, 'page') is True

    books.paginate(page=1, per_page=10)
    assert len(list(books.page.object_list)) == 10

    # new attributes
    assert books.paginator.num_pages == 10
    assert books.page.has_previous() is False
    assert books.page.has_next() is True

    # accessing a non-existant page raises 404
    with pytest.raises(EmptyPage):
        books.paginate(Paginator, page=9999, per_page=10)

    with pytest.raises(PageNotAnInteger):
        books.paginate(Paginator, page='abc', per_page=10)


def test_pagination_shouldnt_prevent_multiple_rendering():
    class SimpleTable(tables.Table):
        name = tables.Column()

    table = SimpleTable([{'name': 'brad'}])
    table.paginate()

    assert table.as_html(request) == table.as_html(request)


def test_empty_text():
    class TestTable(tables.Table):
        a = tables.Column()

    table = TestTable([])
    assert table.empty_text is None

    class TestTable2(tables.Table):
        a = tables.Column()

        class Meta:
            empty_text = 'nothing here'

    table = TestTable2([])
    assert table.empty_text == 'nothing here'

    table = TestTable2([], empty_text='still nothing')
    assert table.empty_text == 'still nothing'


def test_prefix():
    '''
    Test that table prefixes affect the names of querystring parameters
    '''
    class TableA(tables.Table):
        name = tables.Column()

        class Meta:
            prefix = 'x'

    table = TableA([])
    html = table.as_html(build_request('/'))

    assert 'x' == table.prefix
    assert 'xsort=name' in html

    class TableB(tables.Table):
        last_name = tables.Column()

    assert '' == TableB([]).prefix
    assert 'x' == TableB([], prefix='x').prefix

    table = TableB([])
    table.prefix = 'x-'
    html = table.as_html(build_request('/'))

    assert 'x-' == table.prefix
    assert 'x-sort=last_name' in html


def test_field_names():
    class TableA(tables.Table):
        class Meta:
            order_by_field = 'abc'
            page_field = 'def'
            per_page_field = 'ghi'

    table = TableA([])
    assert 'abc' == table.order_by_field
    assert 'def' == table.page_field
    assert 'ghi' == table.per_page_field


def test_field_names_with_prefix():
    class TableA(tables.Table):
        class Meta:
            order_by_field = 'sort'
            page_field = 'page'
            per_page_field = 'per_page'
            prefix = '1-'

    table = TableA([])
    assert '1-sort' == table.prefixed_order_by_field
    assert '1-page' == table.prefixed_page_field
    assert '1-per_page' == table.prefixed_per_page_field

    class TableB(tables.Table):
        class Meta:
            order_by_field = 'sort'
            page_field = 'page'
            per_page_field = 'per_page'

    table = TableB([], prefix='1-')
    assert '1-sort' == table.prefixed_order_by_field
    assert '1-page' == table.prefixed_page_field
    assert '1-per_page' == table.prefixed_per_page_field

    table = TableB([])
    table.prefix = '1-'
    assert '1-sort' == table.prefixed_order_by_field
    assert '1-page' == table.prefixed_page_field
    assert '1-per_page' == table.prefixed_per_page_field


def test_should_support_a_template_to_be_specified():
    class ConstructorSpecifiedTemplateTable(tables.Table):
        name = tables.Column()

    table = ConstructorSpecifiedTemplateTable([], template='dummy.html')
    assert table.template == 'dummy.html'

    class PropertySpecifiedTemplateTable(tables.Table):
        name = tables.Column()

    table = PropertySpecifiedTemplateTable([])
    table.template = 'dummy.html'
    assert table.template == 'dummy.html'

    class DefaultTable(tables.Table):
        pass

    table = DefaultTable([])
    assert table.template == 'django_tables2/table.html'


def test_template_in_meta_class_declaration_should_be_honored():
    class MetaDeclarationSpecifiedTemplateTable(tables.Table):
        name = tables.Column()

        class Meta:
            template = 'dummy.html'

    table = MetaDeclarationSpecifiedTemplateTable([])
    assert table.template == 'dummy.html'
    assert table.as_html(request) == 'dummy template contents\n'


def test_should_support_rendering_multiple_times():
    class MultiRenderTable(tables.Table):
        name = tables.Column()

    # test list data
    table = MultiRenderTable([{'name': 'brad'}])
    assert table.as_html(request) == table.as_html(request)


def test_column_defaults_are_honored():
    class Table(tables.Table):
        name = tables.Column(default='abcd')

        class Meta:
            default = 'efgh'

    table = Table([{}], default='ijkl')
    assert table.rows[0].get_cell('name') == 'abcd'


def test_table_meta_defaults_are_honored():
    class Table(tables.Table):
        name = tables.Column()

        class Meta:
            default = 'abcd'

    table = Table([{}])
    assert table.rows[0].get_cell('name') == 'abcd'


def test_table_defaults_are_honored():
    class Table(tables.Table):
        name = tables.Column()

    table = Table([{}], default='abcd')
    assert table.rows[0].get_cell('name') == 'abcd'

    table = Table([{}], default='abcd')
    table.default = 'efgh'
    assert table.rows[0].get_cell('name') == 'efgh'


AS_VALUES_DATA = [
    {'name': 'Adrian', 'country': 'Australia'},
    {'name': 'Adrian', 'country': 'Brazil'},
    {'name': 'Audrey', 'country': 'Chile'},
    {'name': 'Bassie', 'country': 'Belgium'},
]


def test_as_values():
    class Table(tables.Table):
        name = tables.Column()
        country = tables.Column()

    expected = [['Name', 'Country']] + [[r['name'], r['country']] for r in AS_VALUES_DATA]
    table = Table(AS_VALUES_DATA)

    assert list(table.as_values()) == expected


def test_as_values_exclude():
    class Table(tables.Table):
        name = tables.Column()
        country = tables.Column()

    expected = [['Name']] + [[r['name']] for r in AS_VALUES_DATA]
    table = Table(AS_VALUES_DATA)

    assert list(table.as_values(exclude_columns=('country', ))) == expected


def test_as_values_exclude_from_export():
    class Table(tables.Table):
        name = tables.Column()
        buttons = tables.Column(exclude_from_export=True)

    assert list(Table([]).as_values()) == [['Name'], ]


def test_as_values_empty_values():
    '''
    Table's as_values() method returns `None` for missing values
    '''

    class Table(tables.Table):
        name = tables.Column()
        country = tables.Column()

    data = [
        {'name': 'Adrian', 'country': 'Brazil'},
        {'name': 'Audrey'},
        {'name': 'Bassie', 'country': 'Belgium'},
        {'country': 'France'},
    ]
    expected = [['Name', 'Country']] + [[r.get('name'), r.get('country')] for r in data]
    table = Table(data)
    assert list(table.as_values()) == expected


def test_as_values_render_FOO():

    class Table(tables.Table):
        name = tables.Column()
        country = tables.Column()

        def render_country(self, value):
            return value + ' test'

    expected = [['Name', 'Country']] + [[r['name'], r['country'] + ' test'] for r in AS_VALUES_DATA]

    assert list(Table(AS_VALUES_DATA).as_values()) == expected


def test_as_values_value_FOO():

    class Table(tables.Table):
        name = tables.Column()
        country = tables.Column()

        def render_country(self, value):
            return value + ' test'

        def value_country(self, value):
            return value + ' different'

    expected = [['Name', 'Country']] + [[r['name'], r['country'] + ' different'] for r in AS_VALUES_DATA]

    assert list(Table(AS_VALUES_DATA).as_values()) == expected


def test_row_attrs():
    class Table(tables.Table):
        alpha = tables.Column()
        beta = tables.Column()

    table = Table(MEMORY_DATA, row_attrs={
        'class': lambda record: 'row-id-{}'.format(record['i']),
    })

    assert table.rows[0].attrs == {'class': 'row-id-2 even'}


def test_row_attrs_in_meta():
    class Table(tables.Table):
        alpha = tables.Column()
        beta = tables.Column()

        class Meta:
            row_attrs = {
                'class': lambda record: 'row-id-{}'.format(record['i']),
            }

    table = Table(MEMORY_DATA)
    assert table.rows[0].attrs == {'class': 'row-id-2 even'}


def test_td_attrs_from_table():
    class Table(tables.Table):
        alpha = tables.Column()
        beta = tables.Column()

        class Meta:
            attrs = {
                'td': {
                    'data-column-name': lambda column: column.name
                }
            }
    table = Table(MEMORY_DATA)
    html = table.as_html(request)
    td = parse(html).find('.//tbody/tr[1]/td[1]')
    assert td.attrib == {
        'data-column-name': 'alpha',
        'class': 'alpha'
    }



# coding: utf-8
from __future__ import unicode_literals

import pytest
from django.db import models

import django_tables2 as tables

from ..app.models import Occupation, Person
from ..utils import attrs, build_request


def test_should_be_used_for_booleanfield():
    class BoolModel(models.Model):
        field = models.BooleanField()

        class Meta:
            app_label = 'django_tables2_test'

    class Table(tables.Table):
        class Meta:
            model = BoolModel

    column = Table.base_columns['field']
    assert type(column) == tables.BooleanColumn
    assert column.empty_values != ()


def test_should_be_used_for_nullbooleanfield():
    class NullBoolModel(models.Model):
        field = models.NullBooleanField()

        class Meta:
            app_label = 'django_tables2_test'

    class Table(tables.Table):
        class Meta:
            model = NullBoolModel

    column = Table.base_columns['field']
    assert type(column) == tables.BooleanColumn
    assert column.empty_values == ()


def test_treat_none_different_from_false():
    class Table(tables.Table):
        col = tables.BooleanColumn(null=False, default='---')

    table = Table([{'col': None}])
    assert table.rows[0].get_cell('col') == '---'


def test_treat_none_as_false():
    class Table(tables.Table):
        col = tables.BooleanColumn(null=True)

    table = Table([{'col': None}])
    assert table.rows[0].get_cell('col') == '<span class="false">✘</span>'


def test_value_returns_a_raw_value_without_html():
    class Table(tables.Table):
        col = tables.BooleanColumn()

    table = Table([{'col': True}, {'col': False}])
    assert table.rows[0].get_cell_value('col') == 'True'
    assert table.rows[1].get_cell_value('col') == 'False'


def test_span_attrs():
    class Table(tables.Table):
        col = tables.BooleanColumn(attrs={'span': {'key': 'value'}})

    table = Table([{'col': True}])
    assert attrs(table.rows[0].get_cell('col')) == {'class': 'true', 'key': 'value'}


def test_boolean_field_choices_with_real_model_instances():
    '''
    If a booleanField has choices defined, the value argument passed to
    BooleanColumn.render() is the rendered value, not a bool.
    '''
    class BoolModelChoices(models.Model):
        field = models.BooleanField(choices=(
            (True, 'Yes'),
            (False, 'No'))
        )

        class Meta:
            app_label = 'django_tables2_test'

    class Table(tables.Table):
        class Meta:
            model = BoolModelChoices

    table = Table([BoolModelChoices(field=True), BoolModelChoices(field=False)])

    assert table.rows[0].get_cell('field') == '<span class="true">✔</span>'
    assert table.rows[1].get_cell('field') == '<span class="false">✘</span>'


@pytest.mark.django_db
def test_boolean_field_choices_spanning_relations():
    'The inverse lookup voor boolean choices should also work on related models'

    class Table(tables.Table):
        boolean = tables.BooleanColumn(accessor='occupation.boolean_with_choices')

        class Meta:
            model = Person

    model_true = Occupation.objects.create(
        name='true-name',
        boolean_with_choices=True
    )
    model_false = Occupation.objects.create(
        name='false-name',
        boolean_with_choices=False
    )

    table = Table([
        Person(first_name='True', last_name='False', occupation=model_true),
        Person(first_name='True', last_name='False', occupation=model_false)
    ])

    assert table.rows[0].get_cell('boolean') == '<span class="true">✔</span>'
    assert table.rows[1].get_cell('boolean') == '<span class="false">✘</span>'


@pytest.mark.django_db
def test_boolean_should_not_prevent_rendering_of_other_columns():
    '''Test for issue 360'''
    class Table(tables.Table):
        boolean = tables.BooleanColumn(yesno='waar,onwaar')

        class Meta:
            model = Person
            fields = ('boolean', 'name')

    Occupation.objects.create(name='Waar', boolean=True),
    Occupation.objects.create(name='Onwaar', boolean=False),
    Occupation.objects.create(name='Onduidelijk')

    html = Table(Occupation.objects.all()).as_html(build_request())

    assert 'Waar' in html
    assert 'Onwaar' in html



# coding: utf-8
from __future__ import absolute_import, unicode_literals

import pytest
from django.contrib.auth import get_user_model
from django.template import Context, Template
from django.utils.translation import ugettext_lazy as _

import django_tables2 as tables

from .app.models import Person
from .utils import build_request, parse

User = get_user_model()

data = [
    {'name': 'Adrian', 'country': 'Australia'},
    {'name': 'Roy', 'country': 'Brazil'},
    {'name': 'Audrey', 'country': 'Chile'},
    {'name': 'Bassie', 'country': 'Belgium'},
]


def test_dynamically_adding_columns():
    '''
    When adding columns to self.base_columns, they were actually added to
    the class attribute `Table.base_columns`, and not to the instance
    attribute, `table.base_columns`

    issue #403
    '''
    class MyTable(tables.Table):
        name = tables.Column()

    # this is obvious:
    assert list(MyTable(data).columns.columns.keys()) == ['name']

    assert list(MyTable(data, extra_columns=[
        ('country', tables.Column())
    ]).columns.columns.keys()) == ['name', 'country']

    # this new instance should not have the extra columns added to the first instance.
    assert list(MyTable(data).columns.columns.keys()) == ['name']


def test_sorting_on_dynamically_added_columns():
    class MyTable(tables.Table):
        name = tables.Column()

    table = MyTable(data, order_by='-country', extra_columns=[
        ('country', tables.Column(verbose_name=_('country')))
    ])

    root = parse(table.as_html(build_request()))
    assert root.find('.//tbody/tr/td[2]').text == 'Chile'
    assert root.find('.//tbody/tr[4]/td[2]').text == 'Australia'


@pytest.mark.django_db
def test_dynamically_override_auto_generated_columns():
    for name, country in data:
        Person.objects.create(first_name=name, last_name=country)
    queryset = Person.objects.all()

    class MyTable(tables.Table):
        class Meta:
            model = Person
            fields = ('first_name', 'last_name')

    assert list(MyTable(queryset).columns.columns.keys()) == ['first_name', 'last_name']

    table = MyTable(queryset, extra_columns=[
        ('first_name', tables.Column(attrs={'td': {'style': 'color: red;'}}))
    ])
    # we still should have two columns
    assert list(table.columns.columns.keys()) == ['first_name', 'last_name']
    # the attrs should be applied to the `first_name` column
    assert table.columns['first_name'].attrs['td'] == {'class': 'first_name', 'style': 'color: red;'}


def test_dynamically_add_column_with_sequence():
    class MyTable(tables.Table):
        name = tables.Column()

        class Meta:
            sequence = ('...', 'name')

    assert list(MyTable(data, extra_columns=[
        ('country', tables.Column())
    ]).columns.columns.keys()) == ['country', 'name']

    # override sequence with an argument.
    assert list(MyTable(
        data,
        extra_columns=[('country', tables.Column())],
        sequence=('name', '...')
    ).columns.columns.keys()) == ['name', 'country']


@pytest.mark.django_db
def test_dynamically_hide_columns():
    class MyTable(tables.Table):
        name = tables.Column(orderable=False)
        country = tables.Column(orderable=False)

        def before_render(self, request):
            if request.user.username == 'Bob':
                self.columns.hide('country')
            else:
                self.columns.show('country')

    template = Template('{% load django_tables2 %}{% render_table table %}')

    table = MyTable(data)
    request = build_request(user=User.objects.create(username='Bob'))
    html = table.as_html(request)
    assert '<th class="name">Name</th>' in html
    assert '<th class="country">Country</th>' not in html

    html = template.render(Context({'request': request, 'table': table}))
    assert '<th class="name">Name</th>' in html
    assert '<th class="country">Country</th>' not in html

    request = build_request(user=User.objects.create(username='Alice'))
    html = table.as_html(request)
    assert '<th class="name">Name</th>' in html
    assert '<th class="country">Country</th>' in html

    html = template.render(Context({'request': request, 'table': table}))
    assert '<th class="name">Name</th>' in html
    assert '<th class="country">Country</th>' in html



# coding: utf-8
from __future__ import unicode_literals

import pytest
from django.db import models

import django_tables2 as tables

try:
    from django.contrib.postgres.fields import HStoreField, JSONField
    JSONFIELD_AVAILABLE = True
except ImportError:
    # fields are introduced in django == 1.9
    # remove shim wen we drop support for django 1.8
    JSONFIELD_AVAILABLE = False


@pytest.mark.skipif(not JSONFIELD_AVAILABLE, reason='JSONField added in django 1.9')
def test_should_be_used_for_json_and_hstore_fields():
    class Model(models.Model):
        json = JSONField()
        hstore = HStoreField()

        class Meta:
            app_label = 'django_tables2_test'

    class Table(tables.Table):
        class Meta:
            model = Model

    assert isinstance(Table.base_columns['json'], tables.JSONColumn)
    assert isinstance(Table.base_columns['hstore'], tables.JSONColumn)


def test_jsoncolumn_dict():
    column = tables.JSONColumn()

    record = {'json': {'species': 'Falcon'}}
    html = column.render(value=record['json'], record=record)
    assert html == '<pre >{\n  &quot;species&quot;: &quot;Falcon&quot;\n}</pre>'


def test_jsoncolumn_string():
    column = tables.JSONColumn()

    record = {'json': "really?"}
    html = column.render(value=record['json'], record=record)
    assert html == '<pre >&quot;really?&quot;</pre>'


def test_jsoncolumn_number():
    column = tables.JSONColumn()

    record = {'json': 3.14}
    html = column.render(value=record['json'], record=record)
    assert html == '<pre >3.14</pre>'



# coding: utf-8
from __future__ import unicode_literals

from itertools import count

from django.core.exceptions import ImproperlyConfigured
from django.views.generic.list import ListView

from .config import RequestConfig


class TableMixinBase(object):
    '''
    Base mixin for table-related class based views.
    '''

    context_table_name = 'table'
    table_pagination = None

    def get_table_class(self):
        '''
        Return the class to use for the table.
        '''
        if self.table_class:
            return self.table_class
        klass = type(self).__name__
        raise ImproperlyConfigured(
            'A table class was not specified. Define {}.table_class'.format(klass)
        )

    def get_context_table_name(self, table):
        '''
        Get the name to use for the table's template variable.
        '''
        return self.context_table_name

    def get_table_pagination(self, table):
        '''
        Returns pagination options: True for standard pagination (default),
        False for no pagination, and a dictionary for custom pagination.
        '''
        paginate = self.table_pagination

        if hasattr(self, 'paginate_by') and self.paginate_by is not None:
            # Since ListView knows the concept paginate_by, we use that if no
            # other pagination is configured.
            paginate = paginate or {}
            paginate['per_page'] = self.paginate_by

        if paginate is None:
            return True

        return paginate


class SingleTableMixin(TableMixinBase):
    '''
    Adds a Table object to the context. Typically used with
    `.TemplateResponseMixin`.

    Attributes:
        table_class: subclass of `.Table`
        table_data: data used to populate the table, any compatible data source.
        context_table_name(str): name of the table's template variable (default:
            'table')
        table_pagination (dict): controls table pagination. If a `dict`, passed as
            the *paginate* keyword argument to `.RequestConfig`. As such, any
            Truthy value enables pagination. (default: enable pagination)

    This mixin plays nice with the Django's`.MultipleObjectMixin` by using
    `.get_queryset`` as a fallback for the table data source.
    '''
    table_class = None
    table_data = None

    def get_table(self, **kwargs):
        '''
        Return a table object to use. The table has automatic support for
        sorting and pagination.
        '''
        table_class = self.get_table_class()
        table = table_class(data=self.get_table_data(), **kwargs)
        RequestConfig(self.request, paginate=self.get_table_pagination(table)).configure(table)
        return table

    def get_table_data(self):
        '''
        Return the table data that should be used to populate the rows.
        '''
        if self.table_data is not None:
            return self.table_data
        elif hasattr(self, 'object_list'):
            return self.object_list
        elif hasattr(self, 'get_queryset'):
            return self.get_queryset()

        klass = type(self).__name__
        raise ImproperlyConfigured(
            'Table data was not specified. Define {}.table_data'.format(klass)
        )

    def get_table_kwargs(self):
        '''
        Return the keyword arguments for instantiating the table.

        Allows passing customized arguments to the table constructor, for example, to remove the buttons column,
        you could define this method in your View::

            def get_table_kwargs(self):
                return {
                    'exclude': ('buttons', )
                }
        '''
        return {}

    def get_context_data(self, **kwargs):
        '''
        Overriden version of `.TemplateResponseMixin` to inject the table into
        the template's context.
        '''
        context = super(SingleTableMixin, self).get_context_data(**kwargs)
        table = self.get_table(**self.get_table_kwargs())
        context[self.get_context_table_name(table)] = table
        return context


class SingleTableView(SingleTableMixin, ListView):
    '''
    Generic view that renders a template and passes in a `.Table` instances.
    '''


class MultiTableMixin(TableMixinBase):
    '''
    Adds a Table object to the context. Typically used with
    `.TemplateResponseMixin`.

    the `tables` attribute must be either a list of `.Table` instances or
    classes extended from `.Table` which are not already instantiated. In that
    case, tables_data must be defined, having an entry containing the data for
    each table in `tables`.

    Attributes:
        tables: list of `.Table` instances or list of `.Table` child objects.
        tables_data: if defined, `tables` is assumed to be a list of table
            classes which will be instatiated with the corresponding item from
            this list of `.TableData` instances.
        table_prefix(str): Prefix to be used for each table. The string must
            contain one instance of `{}`, which will be replaced by an integer
            different for each table in the view. Default is 'table_{}-'.
        context_table_name(str): name of the table's template variable (default:
            'tables')

    .. versionadded:: 1.2.3
    '''
    tables = None
    tables_data = None

    table_prefix = 'table_{}-'

    # override context table name to make sense in a multiple table context
    context_table_name = 'tables'

    def get_tables(self):
        if not self.tables:
            klass = type(self).__name__
            raise ImproperlyConfigured(
                'No tables were specified. Define {}.tables'.format(klass)
            )

        return self.tables

    def get_context_data(self, **kwargs):
        context = super(MultiTableMixin, self).get_context_data(**kwargs)

        if self.tables_data is None:
            tables = self.get_tables()
        else:
            data = self.tables_data
            if len(data) != len(self.get_tables()):
                klass = type(self).__name__
                raise ImproperlyConfigured(
                    'len({}.tables_data) != len({}.tables)'.format(klass, klass)
                )
            tables = list(Table(data[i]) for i, Table in enumerate(self.tables))

        # apply prefixes and execute requestConfig for each table
        table_counter = count()
        for table in tables:
            table.prefix = self.table_prefix.format(next(table_counter))

            RequestConfig(self.request, paginate=self.get_table_pagination(table)).configure(table)

            context[self.get_context_table_name(table)] = list(tables)

        return context



# coding: utf-8
from __future__ import unicode_literals

import django_tables2 as tables

from ..utils import attrs


def test_new_attrs_should_be_supported():
    class TestTable(tables.Table):
        col1 = tables.CheckBoxColumn(attrs=dict(th__input={'th_key': 'th_value'},
                                                td__input={'td_key': 'td_value'}))
        col2 = tables.CheckBoxColumn(attrs=dict(input={'key': 'value'}))

    table = TestTable([{'col1': 'data', 'col2': 'data'}])
    assert attrs(table.columns['col1'].header) == {'type': 'checkbox', 'th_key': 'th_value'}
    assert attrs(table.rows[0].get_cell('col1')) == {
        'type': 'checkbox',
        'td_key': 'td_value',
        'value': 'data',
        'name': 'col1'
    }
    assert attrs(table.columns['col2'].header) == {'type': 'checkbox', 'key': 'value'}
    assert attrs(table.rows[0].get_cell('col2')) == {
        'type': 'checkbox',
        'key': 'value',
        'value': 'data',
        'name': 'col2'
    }


def test_column_is_checked():
    class TestTable(tables.Table):
        col = tables.CheckBoxColumn(attrs={'name': 'col'}, checked='is_selected')

    table = TestTable([
        {'col': '1', 'is_selected': True},
        {'col': '2', 'is_selected': False}
    ])
    assert attrs(table.rows[0].get_cell('col')) == {
        'type': 'checkbox',
        'value': '1',
        'name': 'col',
        'checked': 'checked'
    }
    assert attrs(table.rows[1].get_cell('col')) == {'type': 'checkbox', 'value': '2', 'name': 'col'}


def test_column_is_not_checked_for_non_existing_column():
    class TestTable(tables.Table):
        col = tables.CheckBoxColumn(checked='does_not_exist')

    table = TestTable([
        {'col': '1', 'is_selected': True},
        {'col': '2', 'is_selected': False}
    ])
    assert attrs(table.rows[0].get_cell('col')) == {'type': 'checkbox', 'value': '1', 'name': 'col'}
    assert attrs(table.rows[1].get_cell('col')) == {'type': 'checkbox', 'value': '2', 'name': 'col'}


def test_column_is_alway_checked():
    class TestTable(tables.Table):
        col = tables.CheckBoxColumn(checked=True)

    table = TestTable([
        {'col': 1, 'foo': 'bar'},
        {'col': 2, 'foo': 'baz'}
    ])
    assert attrs(table.rows[0].get_cell('col'))['checked'] == 'checked'
    assert attrs(table.rows[1].get_cell('col'))['checked'] == 'checked'


def test_column_is_checked_callback():
    def is_selected(value, record):
        return value == '1'

    class TestTable(tables.Table):
        col = tables.CheckBoxColumn(attrs={'name': 'col'}, checked=is_selected)

    table = TestTable([{'col': '1'}, {'col': '2'}])
    assert attrs(table.rows[0].get_cell('col')) == {
        'type': 'checkbox',
        'value': '1',
        'name': 'col',
        'checked': 'checked'
    }
    assert attrs(table.rows[1].get_cell('col')) == {'type': 'checkbox', 'value': '2', 'name': 'col'}



# coding: utf-8
from __future__ import absolute_import, unicode_literals

import pytest
from django.utils import six

import django_tables2 as tables
from django_tables2.tables import RequestConfig

from .app.models import Person
from .utils import build_request

request = build_request('/')


MEMORY_DATA = [
    {'i': 2, 'alpha': 'b', 'beta': 'b'},
    {'i': 1, 'alpha': 'a', 'beta': 'c'},
    {'i': 3, 'alpha': 'c', 'beta': 'a'},
]


class UnorderedTable(tables.Table):
    i = tables.Column()
    alpha = tables.Column()
    beta = tables.Column()


class OrderedTable(UnorderedTable):
    class Meta:
        order_by = 'alpha'


def test_ordering():
    # fallback to Table.Meta
    assert ('alpha', ) == OrderedTable([], order_by=None).order_by == OrderedTable([]).order_by

    # values of order_by are wrapped in tuples before being returned
    assert OrderedTable([], order_by='alpha').order_by == ('alpha', )
    assert OrderedTable([], order_by=('beta', )).order_by == ('beta', )

    table = OrderedTable([])
    table.order_by = []
    assert () == table.order_by == OrderedTable([], order_by=[]).order_by

    table = OrderedTable([])
    table.order_by = ()
    assert () == table.order_by == OrderedTable([], order_by=()).order_by

    table = OrderedTable([])
    table.order_by = ''
    assert () == table.order_by == OrderedTable([], order_by='').order_by

    # apply an ordering
    table = UnorderedTable([])
    table.order_by = 'alpha'
    assert ('alpha', ) == UnorderedTable([], order_by='alpha').order_by == table.order_by

    table = OrderedTable([])
    table.order_by = 'alpha'
    assert ('alpha', ) == OrderedTable([], order_by='alpha').order_by == table.order_by

    # let's check the data
    table = OrderedTable(MEMORY_DATA, order_by='beta')
    assert 3 == table.rows[0].get_cell('i')

    table = OrderedTable(MEMORY_DATA, order_by='-beta')
    assert 1 == table.rows[0].get_cell('i')

    # allow fallback to Table.Meta.order_by
    table = OrderedTable(MEMORY_DATA)
    assert 1 == table.rows[0].get_cell('i')

    # column's can't be ordered if they're not allowed to be
    class TestTable2(tables.Table):
        a = tables.Column(orderable=False)
        b = tables.Column()

    table = TestTable2([], order_by='a')
    assert table.order_by == ()

    table = TestTable2([], order_by='b')
    assert table.order_by == ('b', )

    # ordering disabled by default
    class TestTable3(tables.Table):
        a = tables.Column(orderable=True)
        b = tables.Column()

        class Meta:
            orderable = False

    table = TestTable3([], order_by='a')
    assert table.order_by == ('a', )

    table = TestTable3([], order_by='b')
    assert table.order_by == ()

    table = TestTable3([], orderable=True, order_by='b')
    assert table.order_by == ('b', )


def test_ordering_different_types():
    from datetime import datetime

    data = [
        {'i': 1, 'alpha': datetime.now(), 'beta': [1]},
        {'i': {}, 'alpha': None, 'beta': ''},
        {'i': 2, 'alpha': None, 'beta': []},
    ]

    table = OrderedTable(data)
    assert '—' == table.rows[0].get_cell('alpha')

    table = OrderedTable(data, order_by='i')
    if six.PY3:
        assert {} == table.rows[0].get_cell('i')
    else:
        assert 1 == table.rows[0].get_cell('i')

    table = OrderedTable(data, order_by='beta')
    assert [] == table.rows[0].get_cell('beta')


brad = {'first_name': 'Bradley', 'last_name': 'Ayers'}
brad2 = {'first_name': 'Bradley', 'last_name': 'Fake'}
chris = {'first_name': 'Chris', 'last_name': 'Doble'}
davina = {'first_name': 'Davina', 'last_name': 'Adisusila'}
ross = {'first_name': 'Ross', 'last_name': 'Ayers'}

people = [brad, brad2, chris, davina, ross]


def test_multi_column_ordering_by_table():

    class PersonTable(tables.Table):
        first_name = tables.Column()
        last_name = tables.Column()

    table = PersonTable(people, order_by=('first_name', 'last_name'))
    assert [brad, brad2, chris, davina, ross] == [r.record for r in table.rows]

    table = PersonTable(people, order_by=('first_name', '-last_name'))
    assert [brad2, brad, chris, davina, ross] == [r.record for r in table.rows]


def test_multi_column_ordering_by_column():
    # let's try column order_by using multiple keys
    class PersonTable(tables.Table):
        name = tables.Column(order_by=('first_name', 'last_name'))

    # add 'name' key for each person.
    for person in people:
        person['name'] = '{p[first_name]} {p[last_name]}'.format(p=person)
    assert brad['name'] == 'Bradley Ayers'

    table = PersonTable(people, order_by='name')
    assert [brad, brad2, chris, davina, ross] == [r.record for r in table.rows]

    table = PersonTable(people, order_by='-name')
    assert [ross, davina, chris, brad2, brad] == [r.record for r in table.rows]


@pytest.mark.django_db
def test_ordering_by_custom_field():
    '''
    When defining a custom field in a table, as name=tables.Column() with
    methods to render and order render_name and order_name, sorting by this
    column causes an error if the custom field is not in last position.
    (issue #413)
    '''

    Person.objects.create(first_name='Alice', last_name='Beta')
    Person.objects.create(first_name='Bob', last_name='Alpha')

    from django.db.models import F, Value
    from django.db.models.functions import Concat

    class PersonTable(tables.Table):
        first_name = tables.Column()
        last_name = tables.Column()
        full_name = tables.Column()

        def render_full_name(self, record):
            return record.last_name + ' ' + record.first_name

        def order_full_name(self, queryset, is_descending):
            queryset = queryset.annotate(
                full_name=Concat(F('last_name'), Value(' '), F('first_name'))
            ).order_by(('-' if is_descending else '') + 'full_name')
            return queryset, True

        class Meta:
            model = Person
            fields = ('first_name', 'last_name', 'full_name')

    table = PersonTable(Person.objects.all())
    request = build_request('/?sort=full_name&sort=first_name')
    RequestConfig(request).configure(table)

    assert table.rows[0].record.first_name == 'Bob'


def test_list_table_data_supports_ordering():
    class Table(tables.Table):
        name = tables.Column()

    data = [
        {'name': 'Bradley'},
        {'name': 'Davina'},
    ]

    table = Table(data)
    assert table.rows[0].get_cell('name') == 'Bradley'
    table.order_by = '-name'
    assert table.rows[0].get_cell('name') == 'Davina'


def test_ordering_non_database_data():
    class Table(tables.Table):
        name = tables.Column()
        country = tables.Column()

    data = [
        {'name': 'Adrian', 'country': 'Australia'},
        {'name': 'Adrian', 'country': 'Brazil'},
        {'name': 'Audrey', 'country': 'Chile'},
        {'name': 'Bassie', 'country': 'Belgium'},
    ]
    table = Table(data, order_by=('-name', '-country'))

    assert table.rows[0].get_cell('name') == 'Bassie'
    assert table.rows[1].get_cell('name') == 'Audrey'
    assert table.rows[2].get_cell('name') == 'Adrian'
    assert table.rows[2].get_cell('country') == 'Brazil'
    assert table.rows[3].get_cell('name') == 'Adrian'
    assert table.rows[3].get_cell('country') == 'Australia'


def test_table_ordering_attributes():
    class Table(tables.Table):
        alpha = tables.Column()
        beta = tables.Column()

    table = Table(MEMORY_DATA, attrs={
        'th': {
            'class': 'custom-header-class',
            '_ordering': {
                'orderable': 'sortable',
                'ascending': 'ascend',
                'descending': 'descend',
            },
        },
    }, order_by='alpha')

    assert 'sortable' in table.columns[0].attrs['th']['class']
    assert 'ascend' in table.columns[0].attrs['th']['class']
    assert 'custom-header-class' in table.columns[1].attrs['th']['class']


def test_table_ordering_attributes_in_meta():
    class Table(tables.Table):
        alpha = tables.Column()
        beta = tables.Column()

        class Meta(OrderedTable.Meta):
            attrs = {
                'th': {
                    'class': 'custom-header-class-in-meta',
                    '_ordering': {
                        'orderable': 'sortable',
                        'ascending': 'ascend',
                        'descending': 'descend',
                    },
                }
            }

    table = Table(MEMORY_DATA)

    assert 'sortable' in table.columns[0].attrs['th']['class']
    assert 'ascend' in table.columns[0].attrs['th']['class']
    assert 'custom-header-class-in-meta' in table.columns[1].attrs['th']['class']


def test_column_ordering_attributes():
    class Table(tables.Table):
        alpha = tables.Column(attrs={
            'th': {
                'class': 'custom-header-class',
                '_ordering': {
                    'orderable': 'sort',
                    'ascending': 'ascending'
                }
            }
        })
        beta = tables.Column(attrs={
            'th': {
                '_ordering': {
                    'orderable': 'canOrder',
                }
            },
            'td': {
                'class': 'cell-2'
            }
        })

    table = Table(MEMORY_DATA, attrs={'class': 'only-on-table'}, order_by='alpha')

    assert 'only-on-table' not in table.columns[0].attrs['th']['class']
    assert 'custom-header-class' in table.columns[0].attrs['th']['class']
    assert 'ascending' in table.columns[0].attrs['th']['class']
    assert 'sort' in table.columns[0].attrs['th']['class']
    assert 'canOrder' in table.columns[1].attrs['th']['class']



# coding: utf-8
from django.conf import settings
from django.conf.urls import include, url
from django.contrib import admin
from django.views import static

from app.views import ClassBased, FilteredPersonListView, MultipleTables, bootstrap, index, multiple, semantic, tutorial

admin.autodiscover()

urlpatterns = [
    url(r'^$', index),
    url(r'^multiple/', multiple, name='multiple'),
    url(r'^class-based/$', ClassBased.as_view(), name='singletableview'),
    url(r'^class-based-multiple/$', MultipleTables.as_view(), name='multitableview'),
    url(r'^class-based-filtered/$', FilteredPersonListView.as_view(), name='filtertableview'),

    url(r'^tutorial/$', tutorial, name='tutorial'),
    url(r'^bootstrap/$', bootstrap, name='bootstrap'),
    url(r'^semantic/$', semantic, name='semantic'),

    url(r'^admin/doc/', include('django.contrib.admindocs.urls')),
    url(r'^admin/', include(admin.site.urls)),

    url(r'^media/(?P<path>.*)$', static.serve, {
        'document_root': settings.MEDIA_ROOT,
    }),
]



from django_filters import FilterSet

from .models import Person


class PersonFilter(FilterSet):
    class Meta:
        model = Person
        fields = {
            'name': ['exact', 'contains'],
            # 'tz': ['exact'],
        }



# coding: utf-8
from __future__ import unicode_literals

from datetime import date

from django.db import models

import django_tables2 as tables


'''
Format string: https://docs.djangoproject.com/en/stable/ref/templates/builtins/#date
D -- Day of the week, textual, 3 letters  -- 'Fri'
b -- Month, textual, 3 letters, lowercase -- 'jan'
Y -- Year, 4 digits.                      -- '1999'
'''


def test_should_handle_explicit_format():
    class TestTable(tables.Table):
        date = tables.DateColumn(format='D b Y')

        class Meta:
            default = '—'

    table = TestTable([{'date': date(2012, 9, 11)},
                       {'date': None}])
    assert table.rows[0].get_cell('date') == 'Tue sep 2012'
    assert table.rows[1].get_cell('date') == '—'


def test_should_handle_long_format(settings):
    settings.DATE_FORMAT = 'D Y b'

    class TestTable(tables.Table):
        date = tables.DateColumn(short=False)

        class Meta:
            default = '—'

    table = TestTable([{'date': date(2012, 9, 11)},
                       {'date': None}])
    assert table.rows[0].get_cell('date') == 'Tue 2012 sep'
    assert table.rows[1].get_cell('date') == '—'


def test_should_handle_short_format(settings):
    settings.SHORT_DATE_FORMAT = 'b Y D'

    class TestTable(tables.Table):
        date = tables.DateColumn(short=True)

        class Meta:
            default = '—'

    table = TestTable([{'date': date(2012, 9, 11)},
                       {'date': None}])
    assert table.rows[0].get_cell('date') == 'sep 2012 Tue'
    assert table.rows[1].get_cell('date') == '—'


def test_should_be_used_for_datefields():
    class DateModel(models.Model):
        field = models.DateField()

        class Meta:
            app_label = 'django_tables2_test'

    class Table(tables.Table):
        class Meta:
            model = DateModel

    assert type(Table.base_columns['field']) == tables.DateColumn


def test_value_returns_a_raw_value_without_html(settings):
    settings.SHORT_DATE_FORMAT = 'b Y D'

    class Table(tables.Table):
        col = tables.DateColumn()

    table = Table([{'col': date(2012, 9, 11)}])
    assert table.rows[0].get_cell_value('col') == 'sep 2012 Tue'



# coding: utf-8
from __future__ import unicode_literals

from datetime import datetime

import pytest
import pytz
from django.db import models

import django_tables2 as tables


'''
Format string: https://docs.djangoproject.com/en/stable/ref/templates/builtins/#date
D -- Day of the week, textual, 3 letters  -- 'Fri'
b -- Month, textual, 3 letters, lowercase -- 'jan'
Y -- Year, 4 digits.                      -- '1999'
A -- 'AM' or 'PM'.                        -- 'AM'
f -- Time, in 12-hour hours[:minutes]     -- '1', '1:30'
'''


@pytest.yield_fixture
def dt():
    dt = datetime(2012, 9, 11, 12, 30, 0)
    yield pytz.timezone('Australia/Brisbane').localize(dt)


def test_should_handle_explicit_format(dt):
    class TestTable(tables.Table):
        date = tables.DateTimeColumn(format='D b Y')

        class Meta:
            default = '—'

    table = TestTable([{'date': dt}, {'date': None}])
    assert table.rows[0].get_cell('date') == 'Tue sep 2012'
    assert table.rows[1].get_cell('date') == '—'


def test_should_handle_long_format(dt, settings):
    class TestTable(tables.Table):
        date = tables.DateTimeColumn(short=False)

        class Meta:
            default = '—'

    settings.DATETIME_FORMAT = 'D Y b A f'
    table = TestTable([{'date': dt}, {'date': None}])
    assert table.rows[0].get_cell('date') == 'Tue 2012 sep PM 12:30'
    assert table.rows[1].get_cell('date') == '—'


def test_should_handle_short_format(dt, settings):
    class TestTable(tables.Table):
        date = tables.DateTimeColumn(short=True)

        class Meta:
            default = '—'

    settings.SHORT_DATETIME_FORMAT = 'b Y D A f'
    table = TestTable([{'date': dt}, {'date': None}])
    assert table.rows[0].get_cell('date') == 'sep 2012 Tue PM 12:30'
    assert table.rows[1].get_cell('date') == '—'


def test_should_be_used_for_datetimefields():
    class DateTimeModel(models.Model):
        field = models.DateTimeField()

        class Meta:
            app_label = 'django_tables2_test'

    class Table(tables.Table):
        class Meta:
            model = DateTimeModel

    assert type(Table.base_columns['field']) == tables.DateTimeColumn


def test_value_returns_a_raw_value_without_html(dt, settings):
    settings.SHORT_DATETIME_FORMAT = 'b Y D A f'

    class Table(tables.Table):
        col = tables.DateTimeColumn()

    table = Table([{'col': dt}])
    assert table.rows[0].get_cell_value('col') == 'sep 2012 Tue PM 12:30'
